#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import Int32Field, Int64Field, AnyField, KeyField
from ...utils import has_unknown_shape
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin
from ..utils import filter_inputs, validate_axis, slice_split, calc_object_length


class TensorDelete(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.DELETE

    _index_obj = AnyField("index_obj")
    _axis = Int32Field("axis")
    _input = KeyField("input")

    # for chunk
    _offset_on_axis = Int64Field("offset_on_axis")

    def __init__(self, index_obj=None, axis=None, offset_on_axis=None, **kw):
        super().__init__(
            _index_obj=index_obj, _axis=axis, _offset_on_axis=offset_on_axis, **kw
        )

    @property
    def index_obj(self):
        return self._index_obj

    @property
    def axis(self):
        return self._axis

    @property
    def offset_on_axis(self):
        return self._offset_on_axis

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if len(self._inputs) > 1:
            self._index_obj = self._inputs[1]

    @classmethod
    def tile(cls, op: "TensorDelete"):
        inp = op.input
        index_obj = op.index_obj
        axis = op.axis
        if axis is None:
            inp = yield from recursive_tile(inp.flatten())
            axis = 0
        if has_unknown_shape(inp):
            yield

        if isinstance(index_obj, int):
            index_obj = [index_obj]

        if isinstance(index_obj, ENTITY_TYPE):
            index_obj = yield from recursive_tile(index_obj.rechunk(index_obj.shape))
            offsets = np.cumsum([0] + list(inp.nsplits[axis]))
            out_chunks = []
            for c in inp.chunks:
                chunk_op = op.copy().reset_key()
                chunk_op._index_obj = index_obj.chunks[0]
                chunk_op._offset_on_axis = int(offsets[c.index[axis]])
                shape = tuple(np.nan if j == axis else s for j, s in enumerate(c.shape))
                out_chunks.append(
                    chunk_op.new_chunk(
                        [c, index_obj.chunks[0]], shape=shape, index=c.index
                    )
                )
            nsplits_on_axis = (np.nan,) * len(inp.nsplits[axis])
        else:
            nsplits_on_axis = [None for _ in inp.nsplits[axis]]
            out_chunks = []
            # index_obj is list, tuple, slice or array like
            if isinstance(index_obj, slice):
                slc_splits = slice_split(index_obj, inp.nsplits[axis])
                for c in inp.chunks:
                    if c.index[axis] in slc_splits:
                        chunk_op = op.copy().reset_key()
                        chunk_slc = slc_splits[c.index[axis]]
                        shape = tuple(
                            s - calc_object_length(chunk_slc, s) if j == axis else s
                            for j, s in enumerate(c.shape)
                        )
                        chunk_op._index_obj = chunk_slc
                        out_chunks.append(
                            chunk_op.new_chunk([c], shape=shape, index=c.index)
                        )
                        nsplits_on_axis[c.index[axis]] = shape[axis]
                    else:
                        out_chunks.append(c)
                        nsplits_on_axis[c.index[axis]] = c.shape[axis]
            else:
                index_obj = np.array(index_obj)
                cum_splits = np.cumsum([0] + list(inp.nsplits[axis]))
                chunk_indexes = defaultdict(list)
                for int_idx in index_obj:
                    in_idx = cum_splits.searchsorted(int_idx, side="right") - 1
                    chunk_indexes[in_idx].append(int_idx - cum_splits[in_idx])

                for c in inp.chunks:
                    idx_on_axis = c.index[axis]
                    if idx_on_axis in chunk_indexes:
                        chunk_op = op.copy().reset_key()
                        chunk_op._index_obj = chunk_indexes[idx_on_axis]
                        shape = tuple(
                            s - len(chunk_indexes[idx_on_axis]) if j == axis else s
                            for j, s in enumerate(c.shape)
                        )
                        out_chunks.append(
                            chunk_op.new_chunk([c], shape=shape, index=c.index)
                        )
                        nsplits_on_axis[c.index[axis]] = shape[axis]
                    else:
                        out_chunks.append(c)
                        nsplits_on_axis[c.index[axis]] = c.shape[axis]

        nsplits = tuple(
            s if i != axis else tuple(nsplits_on_axis)
            for i, s in enumerate(inp.nsplits)
        )
        out = op.outputs[0]
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.input.key]
        index_obj = (
            ctx[op.index_obj.key] if hasattr(op.index_obj, "key") else op.index_obj
        )
        if op.offset_on_axis is None:
            ctx[op.outputs[0].key] = np.delete(inp, index_obj, axis=op.axis)
        else:
            index_obj = np.array(index_obj)
            part_index = [
                idx - op.offset_on_axis
                for idx in index_obj
                if (
                    (idx >= op.offset_on_axis)
                    and idx < (op.offset_on_axis + inp.shape[op.axis or 0])
                )
            ]

            ctx[op.outputs[0].key] = np.delete(inp, part_index, axis=op.axis)

    def __call__(self, arr, obj, shape):
        return self.new_tensor(filter_inputs([arr, obj]), shape=shape, order=arr.order)


def delete(arr, obj, axis=None):
    """
    Return a new array with sub-arrays along an axis deleted. For a one
    dimensional array, this returns those entries not returned by
    `arr[obj]`.

    Parameters
    ----------
    arr : array_like
        Input array.
    obj : slice, int or array of ints
        Indicate indices of sub-arrays to remove along the specified axis.
    axis : int, optional
        The axis along which to delete the subarray defined by `obj`.
        If `axis` is None, `obj` is applied to the flattened array.

    Returns
    -------
    out : mars.tensor
        A copy of `arr` with the elements specified by `obj` removed. Note
        that `delete` does not occur in-place. If `axis` is None, `out` is
        a flattened array.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> arr = mt.array([[1,2,3,4], [5,6,7,8], [9,10,11,12]])
    >>> arr.execute()
    array([[ 1,  2,  3,  4],
           [ 5,  6,  7,  8],
           [ 9, 10, 11, 12]])
    >>> mt.delete(arr, 1, 0).execute()
    array([[ 1,  2,  3,  4],
           [ 9, 10, 11, 12]])
    >>> mt.delete(arr, np.s_[::2], 1).execute()
    array([[ 2,  4],
           [ 6,  8],
           [10, 12]])
    >>> mt.delete(arr, [1,3,5], None).execute()
    array([ 1,  3,  5,  7,  8,  9, 10, 11, 12])
    """
    arr = astensor(arr)
    arr = astensor(arr)
    if getattr(obj, "ndim", 0) > 1:  # pragma: no cover
        raise ValueError(
            "index array argument obj to insert must be one dimensional or scalar"
        )

    if axis is None:
        # if axis is None, array will be flatten
        arr_size = arr.size
        idx_length = calc_object_length(obj, size=arr_size)
        shape = (arr_size - idx_length,)
    else:
        validate_axis(arr.ndim, axis)
        idx_length = calc_object_length(obj, size=arr.shape[axis])
        shape = tuple(
            s - idx_length if i == axis else s for i, s in enumerate(arr.shape)
        )

    op = TensorDelete(index_obj=obj, axis=axis, dtype=arr.dtype)
    return op(arr, obj, shape)



# cython: profile=False, boundscheck=False, wraparound=False, cdivision=True
#
# Licence: BSD 3 clause

# TODO: We still need to use ndarrays instead of typed memoryviews when using
# fused types and when the array may be read-only (for instance when it's
# provided by the user). This is fixed in cython > 0.3.

import numpy as np
cimport numpy as np
from cython cimport floating
from libc.stdlib cimport malloc, free
from libc.float cimport DBL_MAX, FLT_MAX

from ..utils._cython_blas cimport _gemm
from ..utils._cython_blas cimport RowMajor, Trans, NoTrans


np.import_array()


def update_chunk_dense(
        np.ndarray[floating, ndim=2, mode='c'] X,  # IN
        floating[::1] sample_weight,               # IN
        floating[::1] x_squared_norms,             # IN
        floating[:, ::1] centers_old,              # IN
        floating[::1] centers_squared_norms,       # IN
        int[::1] labels,                           # OUT
        floating[:, ::1] centers_new,              # OUT
        floating[::1] weight_in_clusters,          # OUT
        bint update_centers=True):
    cdef:
        int n_samples = X.shape[0]
        int n_clusters = centers_old.shape[0]
        floating *pairwise_distances

    pairwise_distances = <floating*> malloc(n_samples * n_clusters * sizeof(floating))
    result = _update_chunk_dense(&X[0, 0], sample_weight, x_squared_norms,
                                 centers_old, centers_squared_norms,
                                 labels, &centers_new[0, 0],
                                 &weight_in_clusters[0], pairwise_distances,
                                 update_centers)
    free(pairwise_distances)
    return result


cdef void _update_chunk_dense(
        floating *X,                          # IN
        # expecting C aligned 2D array. XXX: Can be
        # replaced by const memoryview when cython min
        # version is >= 0.3
        floating[::1] sample_weight,          # IN
        floating[::1] x_squared_norms,        # IN
        floating[:, ::1] centers_old,         # IN
        floating[::1] centers_squared_norms,  # IN
        int[::1] labels,                      # OUT
        floating *centers_new,                # OUT
        floating *weight_in_clusters,         # OUT
        floating *pairwise_distances,         # OUT
        bint update_centers) nogil:
    """K-means combined EM step for one dense data chunk.

    Compute the partial contribution of a single data chunk to the labels and
    centers.
    """
    cdef:
        int n_samples = labels.shape[0]
        int n_clusters = centers_old.shape[0]
        int n_features = centers_old.shape[1]

        floating sq_dist, min_sq_dist
        int i, j, k, label

    # Instead of computing the full pairwise squared distances matrix,
    # ||X - C||² = ||X||² - 2 X.C^T + ||C||², we only need to store
    # the - 2 X.C^T + ||C||² term since the argmin for a given sample only
    # depends on the centers.
    # pairwise_distances = ||C||²
    for i in range(n_samples):
        for j in range(n_clusters):
            pairwise_distances[i * n_clusters + j] = centers_squared_norms[j]

    # pairwise_distances += -2 * X.dot(C.T)
    _gemm(RowMajor, NoTrans, Trans, n_samples, n_clusters, n_features,
          -2.0, X, n_features, &centers_old[0, 0], n_features,
          1.0, pairwise_distances, n_clusters)

    for i in range(n_samples):
        min_sq_dist = pairwise_distances[i * n_clusters]
        label = 0
        for j in range(1, n_clusters):
            sq_dist = pairwise_distances[i * n_clusters + j]
            if sq_dist < min_sq_dist:
                min_sq_dist = sq_dist
                label = j
        labels[i] = label

        if update_centers:
            weight_in_clusters[label] += sample_weight[i]
            for k in range(n_features):
                centers_new[label * n_features + k] += X[i * n_features + k] * sample_weight[i]


def update_chunk_sparse(
        X,                                         # IN
        floating[::1] sample_weight,               # IN
        floating[::1] x_squared_norms,             # IN
        floating[:, ::1] centers_old,              # IN
        floating[::1] centers_squared_norms,       # IN
        int[::1] labels,                           # OUT
        floating[:, ::1] centers_new,              # OUT
        floating[::1] weight_in_clusters,          # OUT
        bint update_centers=True):
    cdef:
        floating[::1] X_data = X.data
        int[::1] X_indices = X.indices
        int[::1] X_indptr = X.indptr

    X_data = X.data
    X_indices = X.indices
    X_indptr = X.indptr

    return _update_chunk_sparse(
        X_data, X_indices, X_indptr, sample_weight,
        x_squared_norms, centers_old, centers_squared_norms,
        labels, &centers_new[0, 0], &weight_in_clusters[0],
        update_centers
    )


cdef void _update_chunk_sparse(
        floating[::1] X_data,                 # IN
        int[::1] X_indices,                   # IN
        int[::1] X_indptr,                    # IN
        floating[::1] sample_weight,          # IN
        floating[::1] x_squared_norms,        # IN
        floating[:, ::1] centers_old,         # IN
        floating[::1] centers_squared_norms,  # IN
        int[::1] labels,                      # OUT
        floating *centers_new,                # OUT
        floating *weight_in_clusters,         # OUT
        bint update_centers) nogil:
    """K-means combined EM step for one sparse data chunk.

    Compute the partial contribution of a single data chunk to the labels and
    centers.
    """
    cdef:
        int n_samples = labels.shape[0]
        int n_clusters = centers_old.shape[0]
        int n_features = centers_old.shape[1]

        floating sq_dist, min_sq_dist
        int i, j, k, label
        floating max_floating = FLT_MAX if floating is float else DBL_MAX
        int s = X_indptr[0]

    # XXX Precompute the pairwise distances matrix is not worth for sparse
    # currently. Should be tested when BLAS (sparse x dense) matrix
    # multiplication is available.
    for i in range(n_samples):
        min_sq_dist = max_floating
        label = 0

        for j in range(n_clusters):
            sq_dist = 0.0
            for k in range(X_indptr[i] - s, X_indptr[i + 1] - s):
                sq_dist += centers_old[j, X_indices[k]] * X_data[k]

            # Instead of computing the full squared distance with each cluster,
            # ||X - C||² = ||X||² - 2 X.C^T + ||C||², we only need to compute
            # the - 2 X.C^T + ||C||² term since the argmin for a given sample
            # only depends on the centers C.
            sq_dist = centers_squared_norms[j] -2 * sq_dist
            if sq_dist < min_sq_dist:
                min_sq_dist = sq_dist
                label = j

        labels[i] = label

        if update_centers:
            weight_in_clusters[label] += sample_weight[i]
            for k in range(X_indptr[i] - s, X_indptr[i + 1] - s):
                centers_new[label * n_features + X_indices[k]] += X_data[k] * sample_weight[i]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ... import opcodes as OperandDef
from ...config import options
from ...core import recursive_tile
from ...core.operand import OperandStage
from ...lib import sparse
from ...lib.sparse.core import get_array_module as get_sparse_array_module
from ...serialization.serializables import BoolField, Int32Field, Int64Field
from ...utils import has_unknown_shape
from ..operands import TensorMapReduceOperand, TensorOperandMixin, TensorShuffleProxy
from ..array_utils import as_same_device, device
from ..core import TensorOrder
from ..utils import validate_axis, hash_on_axis


class TensorUnique(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.UNIQUE

    _return_index = BoolField("return_index")
    _return_inverse = BoolField("return_inverse")
    _return_counts = BoolField("return_counts")
    _axis = Int32Field("axis")
    _aggregate_size = Int32Field("aggregate_size")

    _start_pos = Int64Field("start_pos")

    def __init__(
        self,
        return_index=None,
        return_inverse=None,
        return_counts=None,
        axis=None,
        start_pos=None,
        aggregate_size=None,
        **kw
    ):
        super().__init__(
            _return_index=return_index,
            _return_inverse=return_inverse,
            _return_counts=return_counts,
            _axis=axis,
            _start_pos=start_pos,
            _aggregate_size=aggregate_size,
            **kw
        )

    @property
    def output_limit(self):
        if self.stage == OperandStage.map:
            return 1
        return (
            1
            + bool(self._return_index)
            + bool(self._return_inverse)
            + bool(self._return_counts)
        )

    @property
    def return_index(self):
        return self._return_index

    @property
    def return_inverse(self):
        return self._return_inverse

    @property
    def return_counts(self):
        return self._return_counts

    @property
    def axis(self):
        return self._axis

    @property
    def aggregate_size(self):
        return self._aggregate_size

    @property
    def start_pos(self):
        return self._start_pos

    @classmethod
    def _gen_kws(cls, op, input_obj, chunk=False, chunk_index=None):
        kws = []

        # unique tensor
        shape = list(input_obj.shape)
        shape[op.axis] = np.nan
        kw = {"shape": tuple(shape), "dtype": input_obj.dtype, "gpu": input_obj.op.gpu}
        if chunk:
            idx = [0] * len(shape)
            idx[op.axis] = chunk_index or 0
            kw["index"] = tuple(idx)
        kws.append(kw)

        # unique indices tensor
        if op.return_index:
            kw = {
                "shape": (np.nan,),
                "dtype": np.dtype(np.intp),
                "gpu": input_obj.op.gpu,
                "type": "indices",
            }
            if chunk:
                kw["index"] = (chunk_index or 0,)
            kws.append(kw)

        # unique inverse tensor
        if op.return_inverse:
            kw = {
                "shape": (input_obj.shape[op.axis],),
                "dtype": np.dtype(np.intp),
                "gpu": input_obj.op.gpu,
                "type": "inverse",
            }
            if chunk:
                kw["index"] = (chunk_index or 0,)
            kws.append(kw)

        # unique counts tensor
        if op.return_counts:
            kw = {
                "shape": (np.nan,),
                "dtype": np.dtype(np.int_),
                "gpu": input_obj.op.gpu,
                "type": "counts",
            }
            if chunk:
                kw["index"] = (chunk_index or 0,)
            kws.append(kw)

        return kws

    def __call__(self, ar):
        from .atleast_1d import atleast_1d

        ar = atleast_1d(ar)
        if self.axis is None:
            if ar.ndim > 1:
                ar = ar.flatten()
            self._axis = 0
        else:
            self._axis = validate_axis(ar.ndim, self._axis)

        kws = self._gen_kws(self, ar)
        tensors = self.new_tensors([ar], kws=kws, order=TensorOrder.C_ORDER)
        if len(tensors) == 1:
            return tensors[0]
        return tensors

    @classmethod
    def _tile_one_chunk(cls, op):
        outs = op.outputs
        ins = op.inputs

        chunk_op = op.copy().reset_key()
        in_chunk = ins[0].chunks[0]
        kws = cls._gen_kws(chunk_op, in_chunk, chunk=True)
        out_chunks = chunk_op.new_chunks([in_chunk], kws=kws, order=outs[0].order)
        new_op = op.copy()
        kws = [out.params.copy() for out in outs]
        for kw, out_chunk in zip(kws, out_chunks):
            kw["chunks"] = [out_chunk]
            kw["nsplits"] = tuple((s,) for s in out_chunk.shape)
        return new_op.new_tensors(ins, kws=kws, order=outs[0].order)

    @classmethod
    def _tile_via_shuffle(cls, op):
        # rechunk the axes except the axis to do unique into 1 chunk
        inp = op.inputs[0]
        if has_unknown_shape(inp):
            yield

        if inp.ndim > 1:
            new_chunk_size = dict()
            for axis in range(inp.ndim):
                if axis == op.axis:
                    continue
                if np.isnan(inp.shape[axis]):
                    yield
                new_chunk_size[axis] = inp.shape[axis]
            if has_unknown_shape(inp):
                yield
            inp = yield from recursive_tile(inp.rechunk(new_chunk_size))

        aggregate_size = op.aggregate_size
        if aggregate_size is None:
            aggregate_size = max(inp.chunk_shape[op.axis] // options.combine_size, 1)

        unique_on_chunk_sizes = inp.nsplits[op.axis]
        start_poses = np.cumsum((0,) + unique_on_chunk_sizes).tolist()[:-1]
        map_chunks = []
        for c in inp.chunks:
            map_op = TensorUnique(
                stage=OperandStage.map,
                return_index=op.return_index,
                return_inverse=op.return_inverse,
                return_counts=op.return_counts,
                axis=op.axis,
                aggregate_size=aggregate_size,
                start_pos=start_poses[c.index[op.axis]],
                dtype=inp.dtype,
            )
            shape = list(c.shape)
            shape[op.axis] = np.nan
            map_chunks.append(map_op.new_chunk([c], shape=tuple(shape), index=c.index))

        shuffle_chunk = TensorShuffleProxy(
            dtype=inp.dtype, _tensor_keys=[inp.op.key]
        ).new_chunk(map_chunks, shape=())

        reduce_chunks = [list() for _ in range(len(op.outputs))]
        for i in range(aggregate_size):
            reduce_op = TensorUnique(
                stage=OperandStage.reduce,
                return_index=op.return_index,
                return_inverse=op.return_inverse,
                return_counts=op.return_counts,
                axis=op.axis,
                reducer_index=(i,),
                reducer_phase="agg",
                reducer_ordinal=i,
                n_reducers=aggregate_size,
            )
            kws = cls._gen_kws(op, inp, chunk=True, chunk_index=i)
            chunks = reduce_op.new_chunks(
                [shuffle_chunk], kws=kws, order=op.outputs[0].order
            )
            for j, c in enumerate(chunks):
                reduce_chunks[j].append(c)

        if op.return_inverse:
            inverse_pos = 2 if op.return_index else 1
            map_inverse_chunks = reduce_chunks[inverse_pos]
            inverse_shuffle_chunk = TensorShuffleProxy(
                dtype=map_inverse_chunks[0].dtype
            ).new_chunk(map_inverse_chunks, shape=())
            inverse_chunks = []
            for j, cs in enumerate(unique_on_chunk_sizes):
                chunk_op = TensorUnique(
                    stage=OperandStage.reduce,
                    reducer_ordinal=j,
                    n_reducers=len(unique_on_chunk_sizes),
                    dtype=map_inverse_chunks[0].dtype,
                    reducer_index=(j,),
                    reducer_phase="inverse",
                )
                inverse_chunk = chunk_op.new_chunk(
                    [inverse_shuffle_chunk], shape=(cs,), index=(j,)
                )
                inverse_chunks.append(inverse_chunk)
            reduce_chunks[inverse_pos] = inverse_chunks

        kws = [out.params for out in op.outputs]
        for kw, chunks in zip(kws, reduce_chunks):
            kw["chunks"] = chunks
        unique_nsplits = list(inp.nsplits)
        unique_nsplits[op.axis] = (np.nan,) * len(reduce_chunks[0])
        kws[0]["nsplits"] = tuple(unique_nsplits)
        i = 1
        if op.return_index:
            kws[i]["nsplits"] = ((np.nan,) * len(reduce_chunks[i]),)
            i += 1
        if op.return_inverse:
            kws[i]["nsplits"] = (inp.nsplits[op.axis],)
            i += 1
        if op.return_counts:
            kws[i]["nsplits"] = ((np.nan,) * len(reduce_chunks[i]),)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, kws=kws)

    @classmethod
    def tile(cls, op: "TensorUnique"):
        if len(op.inputs[0].chunks) == 1:
            return cls._tile_one_chunk(op)
        else:
            return (yield from cls._tile_via_shuffle(op))

    @classmethod
    def _execute_map(cls, ctx, op: "TensorUnique"):
        (ar,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )
        n_reducers = op.aggregate_size

        with device(device_id):
            results = xp.unique(
                ar,
                return_index=op.return_index,
                return_inverse=op.return_inverse,
                return_counts=op.return_counts,
                axis=op.axis,
            )
            results = (results,) if not isinstance(results, tuple) else results
            results_iter = iter(results)
            unique_ar = next(results_iter)
            indices_ar = next(results_iter) + op.start_pos if op.return_index else None
            inverse_ar = next(results_iter) if op.return_inverse else None
            counts_ar = next(results_iter) if op.return_counts else None

            if xp is sparse:
                dense_xp = get_sparse_array_module(unique_ar)
            else:
                dense_xp = xp
            unique_index = (
                dense_xp.arange(unique_ar.shape[op.axis])
                if inverse_ar is not None
                else None
            )
            if unique_ar.size > 0:
                unique_reducers = dense_xp.asarray(
                    hash_on_axis(unique_ar, op.axis, n_reducers)
                )
            else:
                unique_reducers = dense_xp.empty_like(unique_ar)
            ind_ar = dense_xp.arange(ar.shape[op.axis])

            for reducer in range(n_reducers):
                res = []
                cond = unique_reducers == reducer
                # unique
                slc = (slice(None),) * op.axis + (cond,)
                res.append(unique_ar[slc])
                # indices
                if indices_ar is not None:
                    res.append(indices_ar[cond])
                # inverse
                if inverse_ar is not None:
                    index_selected = unique_index[cond]
                    inv_cond = xp.isin(inverse_ar, index_selected)
                    inv_selected = xp.searchsorted(index_selected, inverse_ar[inv_cond])
                    ind_selected = ind_ar[inv_cond]
                    res.append(xp.stack([ind_selected, inv_selected]))
                # counts
                if counts_ar is not None:
                    res.append(counts_ar[cond])
                ctx[op.outputs[0].key, (reducer,)] = tuple(res)

    @classmethod
    def _execute_agg_reduce(cls, ctx, op: "TensorUnique"):
        input_indexes, input_data = zip(*list(op.iter_mapper_data_with_index(ctx)))

        inputs = list(zip(*input_data))
        flatten, device_id, xp = as_same_device(
            list(itertools.chain(*inputs)), device=op.device, ret_extra=True
        )
        n_ret = len(inputs[0])
        inputs = [flatten[i * n_ret : (i + 1) * n_ret] for i in range(len(inputs))]

        inputs_iter = iter(inputs)
        unique_arrays = next(inputs_iter)
        indices_arrays = next(inputs_iter) if op.return_index else None
        inverse_arrays = next(inputs_iter) if op.return_inverse else None
        counts_arrays = next(inputs_iter) if op.return_counts else None

        with device(device_id):
            ar = xp.concatenate(unique_arrays, axis=op.axis)
            result_return_inverse = op.return_inverse or op.return_counts
            axis = op.axis
            if ar.size == 0 or ar.shape[axis] == 0:
                # empty array on the axis
                results = [xp.empty(ar.shape)]
                i = 1
                for it in (op.return_index, op.return_inverse, op.return_counts):
                    if it:
                        results.append(xp.empty([], dtype=op.outputs[i].dtype))
                        i += 1
                results = tuple(results)
            else:
                results = xp.unique(
                    ar,
                    return_index=op.return_index,
                    return_inverse=result_return_inverse,
                    axis=axis,
                )
            results = (results,) if not isinstance(results, tuple) else results
            results_iter = iter(results)
            outputs_iter = iter(op.outputs)
            # unique array
            ctx[next(outputs_iter).key] = next(results_iter)

            if op.output_limit == 1:
                return

            # calc indices
            if op.return_index:
                ctx[next(outputs_iter).key] = xp.concatenate(indices_arrays)[
                    next(results_iter)
                ]
            # calc inverse
            try:
                inverse_result = next(results_iter)
                if op.return_inverse:
                    unique_sizes = tuple(ua.shape[op.axis] for ua in unique_arrays)
                    cum_unique_sizes = np.cumsum((0,) + unique_sizes)
                    indices_out_key = next(outputs_iter).key
                    for i, inverse_array in enumerate(inverse_arrays):
                        p = inverse_result[
                            cum_unique_sizes[i] : cum_unique_sizes[i + 1]
                        ]
                        r = xp.empty(inverse_array.shape, dtype=inverse_array.dtype)
                        if inverse_array.size > 0:
                            r[0] = inverse_array[0]
                            r[1] = p[inverse_array[1]]
                        # return unique length and
                        ctx[indices_out_key, (input_indexes[i][op.axis],)] = (
                            results[0].shape[op.axis],
                            r,
                        )
                # calc counts
                if op.return_counts:
                    result_counts = xp.zeros(results[0].shape[op.axis], dtype=int)
                    t = np.stack([inverse_result, np.concatenate(counts_arrays)])

                    def acc(a):
                        i, v = a
                        result_counts[i] += v

                    np.apply_along_axis(acc, 0, t)
                    ctx[next(outputs_iter).key] = xp.asarray(result_counts)
            except StopIteration:
                pass

    @classmethod
    def _execute_inverse_reduce(cls, ctx, op: "TensorUnique"):
        out = op.outputs[0]
        inputs = list(op.iter_mapper_data(ctx))
        unique_sizes = [inp[0] for inp in inputs]
        cum_unique_sizes = np.cumsum([0] + unique_sizes)
        invs, device_id, xp = as_same_device(
            [inp[1] for inp in inputs], device=op.device, ret_extra=True
        )
        with device(device_id):
            ret = xp.empty(out.shape, dtype=out.dtype)
            for i, inv in enumerate(invs):
                ret[inv[0]] = cum_unique_sizes[i] + inv[1]
            ctx[out.key] = ret

    @classmethod
    def execute(cls, ctx, op: "TensorUnique"):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            if op.reducer_phase == "agg":
                cls._execute_agg_reduce(ctx, op)
            else:
                assert op.reducer_phase == "inverse"
                cls._execute_inverse_reduce(ctx, op)
        else:
            (ar,), device_id, xp = as_same_device(
                [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
            )

            with device(device_id):
                kw = dict(
                    return_index=op.return_index,
                    return_inverse=op.return_inverse,
                    return_counts=op.return_counts,
                )
                if ar.dtype != object and sum(ar.shape) > 0:
                    # axis cannot pass when dtype is object or array size is 0
                    kw["axis"] = op.axis
                results = xp.unique(ar, **kw)
                outs = op.outputs
                if len(outs) == 1:
                    ctx[outs[0].key] = results
                    return

                assert len(outs) == len(results)
                for out, result in zip(outs, results):
                    ctx[out.key] = result


def unique(
    ar,
    return_index=False,
    return_inverse=False,
    return_counts=False,
    axis=None,
    aggregate_size=None,
):
    """
    Find the unique elements of a tensor.

    Returns the sorted unique elements of a tensor. There are three optional
    outputs in addition to the unique elements:

    * the indices of the input tensor that give the unique values
    * the indices of the unique tensor that reconstruct the input tensor
    * the number of times each unique value comes up in the input tensor

    Parameters
    ----------
    ar : array_like
        Input tensor. Unless `axis` is specified, this will be flattened if it
        is not already 1-D.
    return_index : bool, optional
        If True, also return the indices of `ar` (along the specified axis,
        if provided, or in the flattened tensor) that result in the unique tensor.
    return_inverse : bool, optional
        If True, also return the indices of the unique tensor (for the specified
        axis, if provided) that can be used to reconstruct `ar`.
    return_counts : bool, optional
        If True, also return the number of times each unique item appears
        in `ar`.
    axis : int or None, optional
        The axis to operate on. If None, `ar` will be flattened. If an integer,
        the subarrays indexed by the given axis will be flattened and treated
        as the elements of a 1-D tensor with the dimension of the given axis,
        see the notes for more details.  Object tensors or structured tensors
        that contain objects are not supported if the `axis` kwarg is used. The
        default is None.
    aggregate_size: int or None, optional
        How many chunks will be after unique, default as #input.chunks / options.combine_size

    Returns
    -------
    unique : Tensor
        The sorted unique values.
    unique_indices : Tensor, optional
        The indices of the first occurrences of the unique values in the
        original tensor. Only provided if `return_index` is True.
    unique_inverse : Tensor, optional
        The indices to reconstruct the original tensor from the
        unique tensor. Only provided if `return_inverse` is True.
    unique_counts : Tensor, optional
        The number of times each of the unique values comes up in the
        original tensor. Only provided if `return_counts` is True.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.unique([1, 1, 2, 2, 3, 3]).execute()
    array([1, 2, 3])
    >>> a = mt.array([[1, 1], [2, 3]])
    >>> mt.unique(a).execute()
    array([1, 2, 3])

    Return the unique rows of a 2D tensor

    >>> a = mt.array([[1, 0, 0], [1, 0, 0], [2, 3, 4]])
    >>> mt.unique(a, axis=0).execute()
    array([[1, 0, 0], [2, 3, 4]])

    Return the indices of the original tensor that give the unique values:

    >>> a = mt.array(['a', 'b', 'b', 'c', 'a'])
    >>> u, indices = mt.unique(a, return_index=True)
    >>> u.execute()
    array(['a', 'b', 'c'],
           dtype='|S1')
    >>> indices.execute()
    array([0, 1, 3])
    >>> a[indices].execute()
    array(['a', 'b', 'c'],
           dtype='|S1')

    Reconstruct the input array from the unique values:

    >>> a = mt.array([1, 2, 6, 4, 2, 3, 2])
    >>> u, indices = mt.unique(a, return_inverse=True)
    >>> u.execute()
    array([1, 2, 3, 4, 6])
    >>> indices.execute()
    array([0, 1, 4, 3, 1, 2, 1])
    >>> u[indices].execute()
    array([1, 2, 6, 4, 2, 3, 2])
    """
    op = TensorUnique(
        return_index=return_index,
        return_inverse=return_inverse,
        return_counts=return_counts,
        axis=axis,
        aggregate_size=aggregate_size,
    )
    return op(ar)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="binary_or")
class TensorFMod(TensorBinOp):
    _op_type_ = OperandDef.FMOD
    _func_name = "fmod"


@infer_dtype(np.fmod)
def fmod(x1, x2, out=None, where=None, **kwargs):
    """
    Return the element-wise remainder of division.

    This is the NumPy implementation of the C library function fmod, the
    remainder has the same sign as the dividend `x1`. It is equivalent to
    the Matlab(TM) ``rem`` function and should not be confused with the
    Python modulus operator ``x1 % x2``.

    Parameters
    ----------
    x1 : array_like
      Dividend.
    x2 : array_like
      Divisor.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs
        For other keyword-only arguments, see the
        :ref:`ufunc docs <ufuncs.kwargs>`.

    Returns
    -------
    y : Tensor_like
      The remainder of the division of `x1` by `x2`.

    See Also
    --------
    remainder : Equivalent to the Python ``%`` operator.
    divide

    Notes
    -----
    The result of the modulo operation for negative dividend and divisors
    is bound by conventions. For `fmod`, the sign of result is the sign of
    the dividend, while for `remainder` the sign of the result is the sign
    of the divisor. The `fmod` function is equivalent to the Matlab(TM)
    ``rem`` function.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.fmod([-3, -2, -1, 1, 2, 3], 2).execute()
    array([-1,  0, -1,  1,  0,  1])
    >>> mt.remainder([-3, -2, -1, 1, 2, 3], 2).execute()
    array([1, 0, 1, 1, 0, 1])

    >>> mt.fmod([5, 3], [2, 2.]).execute()
    array([ 1.,  1.])
    >>> a = mt.arange(-3, 3).reshape(3, 2)
    >>> a.execute()
    array([[-3, -2],
           [-1,  0],
           [ 1,  2]])
    >>> mt.fmod(a, [2,2]).execute()
    array([[-1,  0],
           [-1,  0],
           [ 1,  0]])
    """
    op = TensorFMod(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ..array_utils import move_to_device
from ..datasource import tensor as astensor
from .core import TensorDeviceConversionBase


class TensorToGPU(TensorDeviceConversionBase):
    _op_type_ = OperandDef.TO_GPU

    def __init__(self, dtype=None, gpu=None, sparse=None, **kw):
        super().__init__(dtype=dtype, gpu=gpu, sparse=sparse, **kw)
        if not self.gpu:
            self.gpu = True

    @classmethod
    def execute(cls, ctx, op):
        device = op.device or 0
        ctx[op.outputs[0].key] = move_to_device(ctx[op.input.key], device)


def to_gpu(x):
    x = astensor(x)

    if x.op.gpu:
        return x

    op = TensorToGPU(dtype=x.dtype)
    return op(x)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob as glob_
import os
from typing import Dict, List
from urllib.parse import urlparse

from ..compression import compress
from .base import path_type, FileSystem
from .local import LocalFileSystem
from .oss import OSSFileSystem


_filesystems = {"file": LocalFileSystem, "oss": OSSFileSystem}


def register_filesystem(name: str, fs):
    _filesystems[name] = fs


def get_fs(path: path_type, storage_options: Dict = None) -> FileSystem:
    if storage_options is None:
        storage_options = dict()

    # detect scheme
    if os.path.exists(path) or glob_.glob(path):
        scheme = "file"
    else:
        scheme = urlparse(path).scheme
    if scheme == "" or len(scheme) == 1:  # len == 1 for windows
        scheme = "file"

    try:
        file_system_type = _filesystems[scheme]
    except KeyError:  # pragma: no cover
        if scheme == "hdfs":
            raise ImportError("Need to install `pyarrow` to connect to HDFS.")
        raise ValueError(
            f"Unknown file system type: {scheme}, "
            f'available include: {", ".join(_filesystems)}'
        )

    if scheme == "file" or scheme == "oss":
        # local file system use an singleton.
        return file_system_type.get_instance()
    else:
        options = file_system_type.parse_from_path(path)
        storage_options.update(options)
        return file_system_type(**storage_options)


def glob(path: path_type, storage_options: Dict = None) -> List[path_type]:
    if "*" in path:
        fs = get_fs(path, storage_options)
        return fs.glob(path)
    else:
        return [path]


def file_size(path: path_type, storage_options: Dict = None) -> int:
    fs = get_fs(path, storage_options)
    return fs.stat(path)["size"]


def open_file(
    path: path_type,
    mode: str = "rb",
    compression: str = None,
    storage_options: Dict = None,
):
    fs = get_fs(path, storage_options)
    file = fs.open(path, mode=mode)

    if compression is not None:
        file = compress(file, compression)

    return file



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Dict

from ....utils import lazy_import, Timer
from ...backend import BaseActorBackend, register_backend
from ..context import MarsActorContext
from .driver import RayActorDriver
from .pool import RayMainPool
from .utils import process_address_to_placement, get_placement_group

ray = lazy_import("ray")

__all__ = ["RayActorBackend"]

logger = logging.getLogger(__name__)


@register_backend
class RayActorBackend(BaseActorBackend):
    @staticmethod
    def name():
        return "ray"

    @staticmethod
    def get_context_cls():
        return MarsActorContext

    @staticmethod
    def get_driver_cls():
        return RayActorDriver

    @classmethod
    async def _create_ray_pools(cls, address: str, n_process: int = None, **kwargs):
        # pop `n_io_process` from kwargs as ray doesn't need this
        kwargs.pop("n_io_process", 0)
        pg_name, bundle_index, _ = process_address_to_placement(address)
        from .pool import RayMainActorPool

        pool_addresses = RayMainActorPool.get_external_addresses(address, n_process)
        assert pool_addresses[0] == address
        pg = get_placement_group(pg_name) if pg_name else None
        num_cpus = kwargs.get("main_pool_cpus", 0)
        sub_pools = {
            sub_pool_address: RayMainActorPool.create_sub_pool(
                address, sub_pool_address
            )
            for sub_pool_address in pool_addresses[1:]
        }
        actor_handle = (
            ray.remote(RayMainPool)
            .options(
                num_cpus=num_cpus,
                name=address,
                max_concurrency=10000000,  # By default, 1000 tasks can be running concurrently.
                max_restarts=-1,  # Auto restarts by ray
                placement_group=pg,
                placement_group_bundle_index=bundle_index,
                placement_group_capture_child_tasks=False,
            )
            .remote(address, n_process, sub_pools, **kwargs)
        )
        pool_handle = RayPoolHandle(actor_handle, sub_pools)
        return pool_handle

    @classmethod
    async def create_actor_pool(cls, address: str, n_process: int = None, **kwargs):
        with Timer() as timer:
            pool_handle = await cls._create_ray_pools(address, n_process, **kwargs)
        logger.info(
            "Submit create actor pool %s took %s seconds.",
            pool_handle.main_pool,
            timer.duration,
        )
        with Timer() as timer:
            await pool_handle.main_pool.start.remote()
        logger.info(
            "Start actor pool %s took %s seconds.",
            pool_handle.main_pool,
            timer.duration,
        )
        return pool_handle


class RayPoolHandle:
    def __init__(
        self,
        main_pool: "ray.actor.ActorHandle",
        sub_pools: Dict[str, "ray.actor.ActorHandle"],
    ):
        self.main_pool = main_pool
        # Hold sub_pool actor handles to avoid gc.
        self.sub_pools = sub_pools

    def __getattr__(self, item):
        if item in ("main_pool", "sub_pools"):  # pragma: no cover
            return object.__getattribute__(self, item)
        return getattr(self.main_pool, item)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...core import OutputType
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameSize(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.REDUCTION_SIZE
    _func_name = "size"

    @property
    def is_atomic(self):
        return True


def size_series(df):
    op = DataFrameSize(output_types=[OutputType.scalar])
    return op(df)


def size_dataframe(df):
    op = DataFrameSize(output_types=[OutputType.series])
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Dict, Any, Type, Union

from ....core import ChunkGraph, Chunk, TileContext
from ....resource import Resource
from ....typing import BandType
from ....utils import merge_dict
from ...subtask import SubtaskGraph, SubtaskResult


class ExecutionConfig:
    """
    The config for execution backends.

    This class should ONLY provide the APIs for the parts other than
    just the execution. Each backend may have a different implementation
    of the API.

    If some configuration is for a specific backend. They should be in
    the backend config. e.g. `get_mars_special_config()` should be in
    the `MarsExecutionConfig`.
    """

    name = None

    def __init__(self, execution_config: Dict):
        """
        An example of execution_config:
        {
            "backend": "mars",
            "mars": {
                "n_worker": 1,
                "n_cpu": 2,
                ...
            },
        }
        """
        self._execution_config = execution_config

    def merge_from(self, execution_config: "ExecutionConfig") -> "ExecutionConfig":
        assert isinstance(execution_config, ExecutionConfig)
        assert self.backend == execution_config.backend
        merge_dict(
            self._execution_config,
            execution_config.get_execution_config(),
        )
        return self

    @property
    def backend(self) -> str:
        """The backend from config."""
        return self._execution_config["backend"]

    def get_execution_config(self) -> Dict:
        """Get the execution config dict."""
        return self._execution_config

    @abstractmethod
    def get_deploy_band_resources(self) -> List[Dict[str, Resource]]:
        """Get the band resources for deployment."""

    @classmethod
    def from_config(cls, config: Dict, backend: str = None) -> "ExecutionConfig":
        """Construct an execution config instance from config."""
        execution_config = config["task"]["execution_config"]
        return cls.from_execution_config(execution_config, backend)

    @classmethod
    def from_execution_config(
        cls, execution_config: Union[Dict, "ExecutionConfig"], backend: str = None
    ) -> "ExecutionConfig":
        """Construct an execution config instance from execution config."""
        if isinstance(execution_config, ExecutionConfig):
            assert backend is None
            return execution_config
        if backend is not None:
            name = execution_config["backend"] = backend
        else:
            name = execution_config.setdefault("backend", "mars")
        config_cls = _name_to_config_cls[name]
        execution_config.setdefault(name, {})
        return config_cls(execution_config)

    @classmethod
    def from_params(
        cls,
        backend: str,
        n_worker: int,
        n_cpu: int,
        mem_bytes: int = 0,
        cuda_devices: List[List[int]] = None,
        **kwargs,
    ) -> "ExecutionConfig":
        """Construct an execution config instance from params."""
        execution_config = {
            "backend": backend,
            backend: dict(
                {
                    "n_worker": n_worker,
                    "n_cpu": n_cpu,
                    "mem_bytes": mem_bytes,
                    "cuda_devices": cuda_devices,
                },
                **kwargs,
            ),
        }
        return cls.from_execution_config(execution_config)


_name_to_config_cls: Dict[str, Type[ExecutionConfig]] = {}


def register_config_cls(config_cls: Type[ExecutionConfig]):
    _name_to_config_cls[config_cls.name] = config_cls
    return config_cls


@dataclass
class ExecutionChunkResult:
    meta: Dict  # The chunk meta for iterative tiling.
    context: Any  # The context info, e.g. ray.ObjectRef.


class TaskExecutor(ABC):
    name = None

    @classmethod
    @abstractmethod
    async def create(
        cls,
        config: Union[Dict, ExecutionConfig],
        *,
        session_id: str,
        address: str,
        task,
        tile_context: TileContext,
        **kwargs,
    ) -> "TaskExecutor":
        backend_config = ExecutionConfig.from_execution_config(config)
        executor_cls = _name_to_task_executor_cls[backend_config.backend]
        if executor_cls.create.__func__ is TaskExecutor.create.__func__:
            raise NotImplementedError(
                f"The {executor_cls} should implement the abstract classmethod `create`."
            )
        return await executor_cls.create(
            backend_config,
            session_id=session_id,
            address=address,
            task=task,
            tile_context=tile_context,
            **kwargs,
        )

    def destroy(self):
        """Destroy the executor."""

    async def __aenter__(self):
        """Called when begin to execute the task."""

    @abstractmethod
    async def execute_subtask_graph(
        self,
        stage_id: str,
        subtask_graph: SubtaskGraph,
        chunk_graph: ChunkGraph,
        tile_context: TileContext,
        context: Any = None,
    ) -> Dict[Chunk, ExecutionChunkResult]:
        """Execute a subtask graph and returns result."""

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Called when finish the task."""

    @abstractmethod
    async def get_available_band_resources(self) -> Dict[BandType, Resource]:
        """Get available band resources."""

    @abstractmethod
    async def get_progress(self) -> float:
        """Get the execution progress."""

    @abstractmethod
    async def cancel(self):
        """Cancel execution."""

    # The following APIs are for compatible with mars backend, they
    # will be removed as soon as possible.
    async def set_subtask_result(self, subtask_result: SubtaskResult):
        """Set the subtask result."""

    def get_stage_processors(self):
        """Get stage processors."""


_name_to_task_executor_cls: Dict[str, Type[TaskExecutor]] = {}


def register_executor_cls(executor_cls: Type[TaskExecutor]):
    _name_to_task_executor_cls[executor_cls.name] = executor_cls
    return executor_cls


class Fetcher:
    """The data fetcher for execution backends."""

    name = None
    required_meta_keys = ()  # The required meta keys.

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    async def append(self, chunk_key: str, chunk_meta: Dict, conditions: List = None):
        """Append chunk key and related infos."""

    @abstractmethod
    async def get(self):
        """Get all the data of appended chunk keys."""

    @classmethod
    def create(cls, backend: str, **kwargs) -> "Fetcher":
        fetcher_cls = _name_to_fetcher_cls[backend]
        return fetcher_cls(**kwargs)


_name_to_fetcher_cls: Dict[str, Type[Fetcher]] = {}


def register_fetcher_cls(fetcher_cls: Type[Fetcher]):
    _name_to_fetcher_cls[fetcher_cls.name] = fetcher_cls
    return fetcher_cls



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from .core import DataFrameDeviceConversionBase


class DataFrameToGPU(DataFrameDeviceConversionBase):
    _op_type_ = OperandDef.TO_GPU

    def __init__(self, dtypes=None, output_types=None, **kw):
        super().__init__(_dtypes=dtypes, _output_types=output_types, **kw)
        if not self.gpu:
            self.gpu = True

    @classmethod
    def execute(cls, ctx, op):
        import cudf

        out_df = op.outputs[0]
        if out_df.ndim == 2:
            ctx[out_df.key] = cudf.DataFrame.from_pandas(ctx[op.inputs[0].key])
        else:
            ctx[out_df.key] = cudf.Series.from_pandas(ctx[op.inputs[0].key])


def to_gpu(df_or_series):
    if df_or_series.op.gpu:
        return df_or_series

    op = DataFrameToGPU()
    return op(df_or_series)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Union

# import aio to ensure patch enabled for Python 3.6
from ..lib import aio

del aio

from . import debug
from .api import (
    actor_ref,
    create_actor,
    has_actor,
    destroy_actor,
    kill_actor,
    Actor,
    StatelessActor,
    create_actor_pool,
    setup_cluster,
    wait_actor_pool_recovered,
    get_pool_config,
)
from .backends import allocate_strategy
from .backends.pool import MainActorPoolType
from .batch import extensible
from .core import ActorRef
from .debug import set_debug_options, get_debug_options, DebugOptions
from .errors import (
    ActorNotExist,
    ActorAlreadyExist,
    ServerClosed,
    SendMessageFailed,
    Return,
)
from .utils import create_actor_ref

# make sure methods are registered
from .backends import mars, ray, test

del mars, ray, test

_T = TypeVar("_T")
ActorRefType = Union[ActorRef, _T]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
from numbers import Integral

import numpy as np

from ..core import TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ..datasource import tensor as astensor
from ..utils import calc_sliced_size, broadcast_shape, replace_ellipsis, index_ndim


_INDEX_ERROR_MSG = (
    "only integers, slices (`:`), ellipsis (`...`), "
    "numpy.newaxis (`None`) and integer or boolean arrays are valid indices"
)


def calc_shape(tensor_shape, index):
    shape = []
    in_axis = 0
    out_axis = 0
    fancy_index = None
    fancy_index_shapes = []
    for ind in index:
        if (
            isinstance(ind, TENSOR_TYPE + TENSOR_CHUNK_TYPE + (np.ndarray,))
            and ind.dtype == np.bool_
        ):
            # bool
            shape.append(np.nan if not isinstance(ind, np.ndarray) else int(ind.sum()))
            for i, t_size, size in zip(
                itertools.count(0),
                ind.shape,
                tensor_shape[in_axis : ind.ndim + in_axis],
            ):
                if not np.isnan(t_size) and not np.isnan(size) and t_size != size:
                    raise IndexError(
                        f"boolean index did not match indexed array along dimension {in_axis + i}; "
                        f"dimension is {size} but corresponding boolean dimension is {t_size}"
                    )
            in_axis += ind.ndim
            out_axis += 1
        elif isinstance(ind, TENSOR_TYPE + TENSOR_CHUNK_TYPE + (np.ndarray,)):
            first_fancy_index = False
            if fancy_index is None:
                first_fancy_index = True
                fancy_index = out_axis
            if isinstance(ind, np.ndarray) and np.any(ind >= tensor_shape[in_axis]):
                out_of_range_index = next(
                    i for i in ind.flat if i >= tensor_shape[in_axis]
                )
                raise IndexError(
                    f"IndexError: index {out_of_range_index} is out of "
                    f"bounds with size {tensor_shape[in_axis]}"
                )
            fancy_index_shapes.append(ind.shape)
            in_axis += 1
            if first_fancy_index:
                out_axis += ind.ndim
        elif isinstance(ind, slice):
            if np.isnan(tensor_shape[in_axis]):
                shape.append(np.nan)
            else:
                shape.append(calc_sliced_size(tensor_shape[in_axis], ind))
            in_axis += 1
            out_axis += 1
        elif isinstance(ind, Integral):
            size = tensor_shape[in_axis]
            if not np.isnan(size) and ind >= size:
                raise IndexError(
                    f"index {ind} is out of bounds for axis {in_axis} with size {size}"
                )
            in_axis += 1
        else:
            assert ind is None
            shape.append(1)

    if fancy_index is not None:
        try:
            if any(np.isnan(np.prod(s)) for s in fancy_index_shapes):
                fancy_index_shape = (np.nan,) * len(fancy_index_shapes[0])
            else:
                fancy_index_shape = broadcast_shape(*fancy_index_shapes)
            shape = shape[:fancy_index] + list(fancy_index_shape) + shape[fancy_index:]
        except ValueError:
            raise IndexError(
                "shape mismatch: indexing arrays could not be broadcast together "
                "with shapes {0}".format(" ".join(str(s) for s in fancy_index_shapes))
            )

    return shape


def preprocess_index(index, convert_bool_to_fancy=None):
    from .nonzero import nonzero

    inds = []
    fancy_indexes = []
    bool_indexes = []
    all_fancy_index_ndarray = True
    all_bool_index_ndarray = True
    for j, ind in enumerate(index):
        if isinstance(ind, (list, np.ndarray) + TENSOR_TYPE):
            if not isinstance(ind, TENSOR_TYPE):
                ind = np.array(ind)
            if ind.dtype.kind not in "biu":
                raise IndexError(_INDEX_ERROR_MSG)
            if ind.dtype.kind == "b":
                # bool indexing
                bool_indexes.append(j)
                if not isinstance(ind, np.ndarray):
                    all_bool_index_ndarray = False
            else:
                # fancy indexing
                fancy_indexes.append(j)
                if not isinstance(ind, np.ndarray):
                    all_fancy_index_ndarray = False
        elif (
            not isinstance(ind, (slice, Integral))
            and ind is not None
            and ind is not Ellipsis
        ):
            raise IndexError(_INDEX_ERROR_MSG)
        inds.append(ind)

    if convert_bool_to_fancy is None:
        convert_bool_to_fancy = (fancy_indexes and len(bool_indexes) > 0) or len(
            bool_indexes
        ) > 1

    if not all_fancy_index_ndarray or (
        convert_bool_to_fancy and not all_bool_index_ndarray
    ):
        # if not all fancy indexes are ndarray,
        # or bool indexes need to be converted to fancy indexes,
        # and not all bool indexes are ndarray,
        # we will convert all of them to Tensor
        for fancy_index in fancy_indexes:
            inds[fancy_index] = astensor(inds[fancy_index])

    # convert bool index to fancy index when any situation below meets:
    # 1. fancy indexes and bool indexes both exists
    # 2. bool indexes more than 2
    if convert_bool_to_fancy:
        default_m = None
        if len(fancy_indexes) > 0:
            default_m = (
                np.nonzero
                if isinstance(inds[fancy_indexes[0]], np.ndarray)
                else nonzero
            )
        for bool_index in bool_indexes:
            ind = inds[bool_index]
            m = default_m
            if m is None:
                m = np.nonzero if isinstance(ind, np.ndarray) else nonzero
            ind = m(ind)[0]
            inds[bool_index] = ind

    return tuple(inds)


def process_index(tensor_ndim, item, convert_bool_to_fancy=None):
    if isinstance(item, list):
        arr = np.array(item)
        if arr.dtype == object:
            item = tuple(item)
        elif arr.dtype.kind == "f":
            raise IndexError(_INDEX_ERROR_MSG)
        else:
            item = (arr,)
    elif not isinstance(item, tuple):
        item = (item,)

    index = preprocess_index(item, convert_bool_to_fancy=convert_bool_to_fancy)
    index = replace_ellipsis(index, tensor_ndim)
    missing = tensor_ndim - sum(index_ndim(i) for i in index)
    if missing < 0:
        raise IndexError("too many indices for tensor")
    return index + (slice(None),) * missing



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .take import take
from .compress import compress
from .extract import extract
from .choose import choose, TensorChoose
from .unravel_index import unravel_index, TensorUnravelIndex
from .nonzero import nonzero, TensorNonzero
from .flatnonzero import flatnonzero
from .slice import TensorSlice
from .getitem import TensorIndex, FancyIndexingDistribute, FancyIndexingConcat
from .setitem import TensorIndexSetValue
from .fill_diagonal import fill_diagonal, TensorFillDiagonal


def _install():
    from ..core import Tensor, TensorData
    from .getitem import _getitem
    from .setitem import _setitem

    setattr(Tensor, "__getitem__", _getitem)
    setattr(TensorData, "__getitem__", _getitem)
    setattr(Tensor, "__setitem__", _setitem)
    setattr(Tensor, "take", take)
    setattr(
        Tensor,
        "compress",
        lambda a, condition, axis=None: compress(condition, a, axis=axis),
    )
    setattr(Tensor, "choose", choose)
    setattr(Tensor, "nonzero", nonzero)


_install()
del _install



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import numpy as np
import scipy.sparse as sp
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import label_binarize as sklearn_label_binarize
from sklearn.utils.sparsefuncs import min_max_axis

from ... import execute as execute_tileable, fetch as fetch_tileable
from ... import opcodes
from ... import tensor as mt
from ...core import ENTITY_TYPE, OutputType, recursive_tile
from ...core.context import get_context, Context
from ...lib.sparse import SparseNDArray
from ...serialization.serializables import AnyField, BoolField, Int32Field, StringField
from ...tensor.core import TensorOrder
from ...typing import TileableType
from ..operands import LearnOperand, LearnOperandMixin
from ..utils import column_or_1d
from ..utils._encode import _unique, _encode
from ..utils.multiclass import unique_labels, type_of_target
from ..utils.validation import _num_samples, check_is_fitted, check_array


class LabelEncoder(TransformerMixin, BaseEstimator):
    """Encode target labels with value between 0 and n_classes-1.

    This transformer should be used to encode target values, *i.e.* `y`, and
    not the input `X`.

    Read more in the :ref:`User Guide <preprocessing_targets>`.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        Holds the label for each class.

    See Also
    --------
    OrdinalEncoder : Encode categorical features using an ordinal encoding
        scheme.
    OneHotEncoder : Encode categorical features as a one-hot numeric array.

    Examples
    --------
    `LabelEncoder` can be used to normalize labels.

    >>> from sklearn import preprocessing
    >>> le = preprocessing.LabelEncoder()
    >>> le.fit([1, 2, 2, 6])
    LabelEncoder()
    >>> le.classes_
    array([1, 2, 6])
    >>> le.transform([1, 1, 2, 6])
    array([0, 0, 1, 2]...)
    >>> le.inverse_transform([0, 0, 1, 2])
    array([1, 1, 2, 6])

    It can also be used to transform non-numerical labels (as long as they are
    hashable and comparable) to numerical labels.

    >>> le = preprocessing.LabelEncoder()
    >>> le.fit(["paris", "paris", "tokyo", "amsterdam"])
    LabelEncoder()
    >>> list(le.classes_)
    ['amsterdam', 'paris', 'tokyo']
    >>> le.transform(["tokyo", "tokyo", "paris"])
    array([2, 2, 1]...)
    >>> list(le.inverse_transform([2, 2, 1]))
    ['tokyo', 'tokyo', 'paris']
    """

    def fit(self, y, session=None, run_kwargs=None, execute=True):
        """Fit label encoder.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : returns an instance of self.
            Fitted label encoder.
        """
        y = column_or_1d(y, warn=True)
        self.classes_ = _unique(y)
        if execute:
            self.classes_ = execute_tileable(
                self.classes_, session=session, **(run_kwargs or dict())
            )
        return self

    def fit_transform(self, y, session=None, run_kwargs=None):
        """Fit label encoder and return encoded labels.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        y : array-like of shape (n_samples,)
            Encoded labels.
        """
        y = column_or_1d(y, warn=True)
        self.classes_, y = execute_tileable(
            _unique(y, return_inverse=True), session=session, **(run_kwargs or dict())
        )
        return y

    def transform(self, y, session=None, run_kwargs=None, execute=True):
        """Transform labels to normalized encoding.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        y : array-like of shape (n_samples,)
            Labels as normalized encodings.
        """
        check_is_fitted(self)
        y = column_or_1d(y, warn=True)
        # transform of empty array is empty array
        if _num_samples(y) == 0:
            return mt.array([])

        t = _encode(y, uniques=self.classes_)
        if execute:
            t = t.execute(session=session, **(run_kwargs or dict()))
        return t

    def inverse_transform(self, y, session=None, run_kwargs=None):
        """Transform labels back to original encoding.

        Parameters
        ----------
        y : ndarray of shape (n_samples,)
            Target values.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            Original encoding.
        """
        check_is_fitted(self)
        y = column_or_1d(y, warn=True)
        # inverse transform of empty array is empty array
        if _num_samples(y) == 0:
            return mt.array([])

        def _class_checker(chunk_data, classes_data):
            diff = np.setdiff1d(chunk_data, np.arange(len(classes_data)))
            if len(diff):
                raise ValueError("y contains previously unseen labels: %s" % str(diff))
            return chunk_data

        y = mt.asarray(y).map_chunk(_class_checker, args=(self.classes_,))
        return self.classes_[y].execute(session=session, **(run_kwargs or dict()))

    def _more_tags(self):
        return {"X_types": ["1dlabels"]}


class LabelBinarizer(TransformerMixin, BaseEstimator):
    """Binarize labels in a one-vs-all fashion.

    Several regression and binary classification algorithms are
    available in scikit-learn. A simple way to extend these algorithms
    to the multi-class classification case is to use the so-called
    one-vs-all scheme.

    At learning time, this simply consists in learning one regressor
    or binary classifier per class. In doing so, one needs to convert
    multi-class labels to binary labels (belong or does not belong
    to the class). LabelBinarizer makes this process easy with the
    transform method.

    At prediction time, one assigns the class for which the corresponding
    model gave the greatest confidence. LabelBinarizer makes this easy
    with the inverse_transform method.

    Read more in the :ref:`User Guide <preprocessing_targets>`.

    Parameters
    ----------

    neg_label : int, default=0
        Value with which negative labels must be encoded.

    pos_label : int, default=1
        Value with which positive labels must be encoded.

    sparse_output : bool, default=False
        True if the returned array from transform is desired to be in sparse
        CSR format.

    Attributes
    ----------

    classes_ : ndarray of shape (n_classes,)
        Holds the label for each class.

    y_type_ : str
        Represents the type of the target data as evaluated by
        utils.multiclass.type_of_target. Possible type are 'continuous',
        'continuous-multioutput', 'binary', 'multiclass',
        'multiclass-multioutput', 'multilabel-indicator', and 'unknown'.

    sparse_input_ : bool
        True if the input data to transform is given as a sparse matrix, False
        otherwise.

    Examples
    --------
    >>> from mars.learn import preprocessing
    >>> lb = preprocessing.LabelBinarizer()
    >>> lb.fit([1, 2, 6, 4, 2])
    LabelBinarizer()
    >>> lb.classes_
    array([1, 2, 4, 6])
    >>> lb.transform([1, 6])
    array([[1, 0, 0, 0],
           [0, 0, 0, 1]])

    Binary targets transform to a column vector

    >>> lb = preprocessing.LabelBinarizer()
    >>> lb.fit_transform(['yes', 'no', 'no', 'yes'])
    array([[1],
           [0],
           [0],
           [1]])

    Passing a 2D matrix for multilabel classification

    >>> import numpy as np
    >>> lb.fit(np.array([[0, 1, 1], [1, 0, 0]]))
    LabelBinarizer()
    >>> lb.classes_
    array([0, 1, 2])
    >>> lb.transform([0, 1, 2, 1])
    array([[1, 0, 0],
           [0, 1, 0],
           [0, 0, 1],
           [0, 1, 0]])

    See Also
    --------
    label_binarize : Function to perform the transform operation of
        LabelBinarizer with fixed classes.
    OneHotEncoder : Encode categorical features using a one-hot aka one-of-K
        scheme.
    """

    def __init__(self, *, neg_label=0, pos_label=1, sparse_output=False):
        if neg_label >= pos_label:
            raise ValueError(
                "neg_label={0} must be strictly less than "
                "pos_label={1}.".format(neg_label, pos_label)
            )

        if sparse_output and (pos_label == 0 or neg_label != 0):
            raise ValueError(
                "Sparse binarization is only supported with non "
                "zero pos_label and zero neg_label, got "
                "pos_label={0} and neg_label={1}"
                "".format(pos_label, neg_label)
            )

        self.neg_label = neg_label
        self.pos_label = pos_label
        self.sparse_output = sparse_output

    def fit(self, y, session=None, run_kwargs=None):
        """Fit label binarizer.

        Parameters
        ----------
        y : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Target values. The 2-d matrix should only contain 0 and 1,
            represents multilabel classification.

        Returns
        -------
        self : returns an instance of self.
        """
        self.y_type_ = fetch_tileable(
            execute_tileable(
                type_of_target(y), session=session, **(run_kwargs or dict())
            )
        )
        if "multioutput" in self.y_type_:
            raise ValueError(
                "Multioutput target data is not supported with label binarization"
            )
        if _num_samples(y) == 0:  # pragma: no cover
            raise ValueError("y has 0 samples: %r" % y)

        self.sparse_input_ = mt.tensor(y).issparse()
        self.classes_ = unique_labels(y).execute(
            session=session, **(run_kwargs or dict())
        )
        return self

    def fit_transform(self, y, session=None, run_kwargs=None):
        """Fit label binarizer and transform multi-class labels to binary
        labels.

        The output of transform is sometimes referred to as
        the 1-of-K coding scheme.

        Parameters
        ----------
        y : {ndarray, sparse matrix} of shape (n_samples,) or \
                (n_samples, n_classes)
            Target values. The 2-d matrix should only contain 0 and 1,
            represents multilabel classification. Sparse matrix can be
            CSR, CSC, COO, DOK, or LIL.

        Returns
        -------
        Y : {ndarray, sparse matrix} of shape (n_samples, n_classes)
            Shape will be (n_samples, 1) for binary problems. Sparse matrix
            will be of CSR format.
        """
        return self.fit(y, session=session, run_kwargs=run_kwargs).transform(
            y, session=session, run_kwargs=run_kwargs
        )

    def transform(self, y, session=None, run_kwargs=None):
        """Transform multi-class labels to binary labels.

        The output of transform is sometimes referred to by some authors as
        the 1-of-K coding scheme.

        Parameters
        ----------
        y : {array, sparse matrix} of shape (n_samples,) or \
                (n_samples, n_classes)
            Target values. The 2-d matrix should only contain 0 and 1,
            represents multilabel classification. Sparse matrix can be
            CSR, CSC, COO, DOK, or LIL.

        Returns
        -------
        Y : {ndarray, sparse matrix} of shape (n_samples, n_classes)
            Shape will be (n_samples, 1) for binary problems. Sparse matrix
            will be of CSR format.
        """
        check_is_fitted(self)

        target = fetch_tileable(
            execute_tileable(
                type_of_target(y), session=session, **(run_kwargs or dict())
            )
        )
        y_is_multilabel = target.startswith("multilabel")
        if y_is_multilabel and not self.y_type_.startswith("multilabel"):
            raise ValueError("The object was not fitted with multilabel input.")

        return label_binarize(
            y,
            classes=self.classes_,
            pos_label=self.pos_label,
            neg_label=self.neg_label,
            sparse_output=self.sparse_output,
        )

    def inverse_transform(self, Y, threshold=None):
        """Transform binary labels back to multi-class labels.

        Parameters
        ----------
        Y : {ndarray, sparse matrix} of shape (n_samples, n_classes)
            Target values. All sparse matrices are converted to CSR before
            inverse transformation.

        threshold : float, default=None
            Threshold used in the binary and multi-label cases.

            Use 0 when ``Y`` contains the output of decision_function
            (classifier).
            Use 0.5 when ``Y`` contains the output of predict_proba.

            If None, the threshold is assumed to be half way between
            neg_label and pos_label.

        Returns
        -------
        y : {ndarray, sparse matrix} of shape (n_samples,)
            Target values. Sparse matrix will be of CSR format.

        Notes
        -----
        In the case when the binary labels are fractional
        (probabilistic), inverse_transform chooses the class with the
        greatest value. Typically, this allows to use the output of a
        linear model's decision_function method directly as the input
        of inverse_transform.
        """
        check_is_fitted(self)

        if threshold is None:
            threshold = (self.pos_label + self.neg_label) / 2.0

        Y = mt.asarray(Y)
        if self.y_type_ == "multiclass":
            y_inv = Y.map_chunk(
                _inverse_binarize_multiclass,
                args=(self.classes_,),
                dtype=self.classes_.dtype,
                shape=(Y.shape[0],),
            )
        else:
            shape = (Y.shape[0],) if self.y_type_ != "multilabel-indicator" else Y.shape
            y_inv = Y.map_chunk(
                _inverse_binarize_thresholding,
                args=(self.y_type_, self.classes_, threshold),
                dtype=self.classes_.dtype,
                shape=shape,
            )

        if self.sparse_input_:
            y_inv = y_inv.tosparse()
        elif y_inv.issparse():
            y_inv = y_inv.todense()

        return y_inv

    def _more_tags(
        self,
    ):  # pragma: no cover  # noqa: R0201  # pylint: disable=no-self-use
        return {"X_types": ["1dlabels"]}


class LabelBinarize(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.LABEL_BINARIZE

    y = AnyField("y")
    classes = AnyField("classes")
    neg_label = Int32Field("neg_label")
    pos_label = Int32Field("pos_label")
    sparse_output = BoolField("sparse_output")
    # for chunk
    y_type = StringField("y_type")
    pos_switch = BoolField("pos_switch")

    def __call__(self, y: TileableType, classes: TileableType):
        inputs = []
        if not isinstance(y, list):
            # XXX Workaround that will be removed when list of list format is
            # dropped
            self.y = y = check_array(y, accept_sparse=True, ensure_2d=False, dtype=None)
        if isinstance(y, ENTITY_TYPE):
            inputs.append(y)
        if isinstance(classes, ENTITY_TYPE):
            inputs.append(classes)
        self.sparse = self.sparse_output
        self.output_types = [OutputType.tensor]
        if len(classes) == 2:
            n_dim1 = 1
        else:
            n_dim1 = len(classes)
        return self.new_tileable(
            inputs,
            shape=(np.nan, n_dim1),
            dtype=np.dtype(int),
            order=TensorOrder.C_ORDER,
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if isinstance(self.y, ENTITY_TYPE):
            self.y = self._inputs[0]
        if isinstance(self.classes, ENTITY_TYPE):
            self.classes = self._inputs[-1]

    @classmethod
    def tile(cls, op: "LabelBinarize"):
        y = op.y
        classes = op.classes
        neg_label = op.neg_label
        pos_label = op.pos_label
        sparse_output = op.sparse_output
        out = op.outputs[0]
        ctx = get_context()

        if isinstance(y, list):
            if _num_samples(y) == 0:
                raise ValueError("y has 0 samples: %r" % y)

        if len(op.inputs) == 0:
            # no entity input
            r = sklearn_label_binarize(
                op.y,
                classes=op.classes,
                neg_label=neg_label,
                pos_label=pos_label,
                sparse_output=sparse_output,
            )
            return (yield from recursive_tile(mt.tensor(r)))
        else:
            # trigger execution
            yield

        if neg_label >= pos_label:
            raise ValueError(
                "neg_label={0} must be strictly less than "
                "pos_label={1}.".format(neg_label, pos_label)
            )

        if sparse_output and (pos_label == 0 or neg_label != 0):
            raise ValueError(
                "Sparse binarization is only supported with non "
                "zero pos_label and zero neg_label, got "
                "pos_label={0} and neg_label={1}"
                "".format(pos_label, neg_label)
            )

        # To account for pos_label == 0 in the dense case
        pos_switch = pos_label == 0
        if pos_switch:
            pos_label = -neg_label

        y_type = yield from recursive_tile(type_of_target(y))
        yield y_type.chunks
        y_type = ctx.get_chunks_result([y_type.chunks[0].key])[0]
        y_type = y_type.item() if hasattr(y_type, "item") else y_type
        if "multioutput" in y_type:
            raise ValueError(
                "Multioutput target data is not supported with label binarization"
            )
        if y_type == "unknown":
            raise ValueError("The type of target data is not known")

        n_samples = mt.tensor(y).shape[0]
        n_classes = len(classes)

        if y_type == "binary":
            if n_classes == 1:
                if sparse_output:
                    return (
                        yield from recursive_tile(
                            mt.zeros((n_samples, 1), dtype=int, sparse=True)
                        )
                    )
                else:
                    Y = mt.zeros((len(y), 1), dtype=int)
                    Y += neg_label
                    return (yield from recursive_tile(Y))
            elif len(classes) >= 3:
                y_type = "multiclass"

        if y_type == "multilabel-indicator":
            y_n_classes = y.shape[1] if hasattr(y, "shape") else len(y[0])
            if mt.tensor(classes).size != y_n_classes:
                raise ValueError(
                    "classes {0} mismatch with the labels {1}"
                    " found in the data".format(classes, unique_labels(y))
                )

        if y_type in ("binary", "multiclass"):
            y = yield from recursive_tile(column_or_1d(y))
            if y_type == "binary":
                out_shape = (n_samples, 1)
            else:
                out_shape = (n_samples, n_classes)
        elif y_type == "multilabel-indicator":
            out_shape = y.shape
        else:
            raise ValueError(
                "%s target data is not supported with label binarization" % y_type
            )

        out_chunks = []
        for y_chunk in y.chunks:
            chunk_inputs = [y_chunk]
            classes_chunk = classes
            if isinstance(classes, ENTITY_TYPE):
                chunk_inputs.append(classes.chunks[0])
                classes_chunk = classes.chunks[0]
            chunk_op = LabelBinarize(
                y=y_chunk,
                classes=classes_chunk,
                neg_label=neg_label,
                pos_label=pos_label,
                sparse_output=sparse_output,
                y_type=y_type,
                pos_switch=pos_switch,
                _output_types=op.output_types,
            )
            if len(out_shape) == 2:
                chunk_shape = (y_chunk.shape[0], out_shape[1])
                chunk_index = (y_chunk.index[0], 0)
            else:  # pragma: no cover
                chunk_shape = (y_chunk.shape[0],)
                chunk_index = (y_chunk.index[0],)
            out_chunk = chunk_op.new_chunk(
                chunk_inputs,
                shape=chunk_shape,
                dtype=out.dtype,
                order=out.order,
                index=chunk_index,
            )
            out_chunks.append(out_chunk)

        params = out.params.copy()
        params["chunks"] = out_chunks
        params["shape"] = out_shape
        if len(out_shape) == 2:
            nsplits = (y.nsplits[0], (out_shape[1],))
        else:  # pragma: no cover
            nsplits = (y.nsplits[0],)
        params["nsplits"] = nsplits
        return op.copy().new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "LabelBinarize"):
        y = ctx[op.y.key]
        if hasattr(y, "raw"):
            # SparseNDArray
            y = y.raw
        if isinstance(op.classes, ENTITY_TYPE):
            classes = ctx[op.classes.key]
        else:
            classes = op.classes
        y_type = op.y_type
        sparse_output = op.sparse_output
        pos_label = op.pos_label
        neg_label = op.neg_label
        pos_switch = op.pos_switch

        n_samples = y.shape[0] if sp.issparse(y) else len(y)
        n_classes = len(classes)
        sorted_class = np.sort(classes)

        if y_type in ("binary", "multiclass"):
            # pick out the known labels from y
            y_in_classes = np.in1d(y, classes)
            y_seen = y[y_in_classes]
            indices = np.searchsorted(sorted_class, y_seen)
            indptr = np.hstack((0, np.cumsum(y_in_classes)))

            data = np.empty_like(indices)
            data.fill(pos_label)
            Y = sp.csr_matrix((data, indices, indptr), shape=(n_samples, n_classes))
        elif y_type == "multilabel-indicator":
            Y = sp.csr_matrix(y)
            if pos_label != 1:
                data = np.empty_like(Y.data)
                data.fill(pos_label)
                Y.data = data
        else:  # pragma: no cover
            raise ValueError(
                "%s target data is not supported with label binarization" % y_type
            )

        if not sparse_output:
            Y = Y.toarray()
            Y = Y.astype(int, copy=False)

            if neg_label != 0:
                Y[Y == 0] = neg_label

            if pos_switch:
                Y[Y == pos_label] = 0
        else:
            Y.data = Y.data.astype(int, copy=False)

        # preserve label ordering
        if np.any(classes != sorted_class):
            indices = np.searchsorted(sorted_class, classes)
            Y = Y[:, indices]

        if y_type == "binary":
            if sparse_output:
                Y = Y.getcol(-1)
            else:
                Y = Y[:, -1].reshape((-1, 1))

        if sp.issparse(Y):
            Y = SparseNDArray(Y)
        ctx[op.outputs[0].key] = Y


def label_binarize(
    y, *, classes, neg_label=0, pos_label=1, sparse_output=False, execute=True
):
    """Binarize labels in a one-vs-all fashion.

    Several regression and binary classification algorithms are
    available in scikit-learn. A simple way to extend these algorithms
    to the multi-class classification case is to use the so-called
    one-vs-all scheme.

    This function makes it possible to compute this transformation for a
    fixed set of class labels known ahead of time.

    Parameters
    ----------
    y : array-like
        Sequence of integer labels or multilabel data to encode.

    classes : array-like of shape (n_classes,)
        Uniquely holds the label for each class.

    neg_label : int, default=0
        Value with which negative labels must be encoded.

    pos_label : int, default=1
        Value with which positive labels must be encoded.

    sparse_output : bool, default=False,
        Set to true if output binary array is desired in CSR sparse format.

    Returns
    -------
    Y : {tensor, sparse tensor} of shape (n_samples, n_classes)
        Shape will be (n_samples, 1) for binary problems.

    Examples
    --------
    >>> from mars.learn.preprocessing import label_binarize
    >>> label_binarize([1, 6], classes=[1, 2, 4, 6])
    array([[1, 0, 0, 0],
           [0, 0, 0, 1]])

    The class ordering is preserved:

    >>> label_binarize([1, 6], classes=[1, 6, 4, 2])
    array([[1, 0, 0, 0],
           [0, 1, 0, 0]])

    Binary targets transform to a column vector

    >>> label_binarize(['yes', 'no', 'no', 'yes'], classes=['no', 'yes'])
    array([[1],
           [0],
           [0],
           [1]])

    See Also
    --------
    LabelBinarizer : Class used to wrap the functionality of label_binarize and
        allow for fitting to classes independently of the transform operation.
    """
    op = LabelBinarize(
        y=y,
        classes=classes,
        neg_label=neg_label,
        pos_label=pos_label,
        sparse_output=sparse_output,
    )
    result = op(y, classes)
    return result.execute() if execute else result


def _inverse_binarize_multiclass(y, classes):  # pragma: no cover
    """Inverse label binarization transformation for multiclass.

    Multiclass uses the maximal score instead of a threshold.
    """
    classes = np.asarray(classes)

    if sp.issparse(y):
        # Find the argmax for each row in y where y is a CSR matrix

        y = y.tocsr()
        n_samples, n_outputs = y.shape
        outputs = np.arange(n_outputs)
        row_max = min_max_axis(y, 1)[1]
        row_nnz = np.diff(y.indptr)

        y_data_repeated_max = np.repeat(row_max, row_nnz)
        # picks out all indices obtaining the maximum per row
        y_i_all_argmax = np.flatnonzero(y_data_repeated_max == y.data)

        # For corner case where last row has a max of 0
        if row_max[-1] == 0:
            y_i_all_argmax = np.append(y_i_all_argmax, [len(y.data)])

        # Gets the index of the first argmax in each row from y_i_all_argmax
        index_first_argmax = np.searchsorted(y_i_all_argmax, y.indptr[:-1])
        # first argmax of each row
        y_ind_ext = np.append(y.indices, [0])
        y_i_argmax = y_ind_ext[y_i_all_argmax[index_first_argmax]]
        # Handle rows of all 0
        y_i_argmax[np.where(row_nnz == 0)[0]] = 0

        # Handles rows with max of 0 that contain negative numbers
        samples = np.arange(n_samples)[(row_nnz > 0) & (row_max.ravel() == 0)]
        for i in samples:
            ind = y.indices[y.indptr[i] : y.indptr[i + 1]]
            y_i_argmax[i] = classes[np.setdiff1d(outputs, ind)][0]

        return classes[y_i_argmax]
    else:
        return classes.take(y.argmax(axis=1), mode="clip")


def _inverse_binarize_thresholding(
    y, output_type, classes, threshold
):  # pragma: no cover
    """Inverse label binarization transformation using thresholding."""

    if output_type == "binary" and y.ndim == 2 and y.shape[1] > 2:
        raise ValueError("output_type='binary', but y.shape = {0}".format(y.shape))

    if output_type != "binary" and y.shape[1] != len(classes):
        raise ValueError(
            "The number of class is not equal to the number of dimension of y."
        )

    classes = np.asarray(classes)

    # Perform thresholding
    if sp.issparse(y):
        if threshold > 0:
            if y.format not in ("csr", "csc"):
                y = y.tocsr()
            y.data = np.array(y.data > threshold, dtype=int)
            y.eliminate_zeros()
        else:
            y = np.array(y.toarray() > threshold, dtype=int)
    else:
        y = np.array(y > threshold, dtype=int)

    # Inverse transform data
    if output_type == "binary":
        if sp.issparse(y):
            y = y.toarray()
        if y.ndim == 2 and y.shape[1] == 2:
            return classes[y[:, 1]]
        else:
            if len(classes) == 1:
                return np.repeat(classes[0], len(y))
            else:
                return classes[y.ravel()]

    elif output_type == "multilabel-indicator":
        return y

    else:
        raise ValueError("{0} format is not supported".format(output_type))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Any, Callable, List

from .... import oscar as mo
from ....storage import StorageLevel
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import DataInfo
from .core import AbstractStorageAPI


class StorageWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = "/api/session/(?P<session_id>[^/]+)/storage"

    async def _get_oscar_meta_api(self, session_id: str):
        from ...meta import MetaAPI

        return await self._get_api_by_key(MetaAPI, session_id)

    async def _get_storage_api_by_object_id(self, session_id: str, object_id: str):
        from .oscar import StorageAPI

        meta_api = await self._get_oscar_meta_api(session_id)
        bands = (await meta_api.get_chunk_meta(object_id, ["bands"])).get("bands")
        if not bands:
            raise KeyError
        return await StorageAPI.create(session_id, bands[0][0], bands[0][1])

    @web_api("(?P<data_key>[^/]+)", method="get")
    async def get_data(self, session_id: str, data_key: str):
        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        result = await oscar_api.get(data_key)
        self.write(serialize_serializable(result))

    @web_api("batch/get", method="post")
    async def get_batch_data(self, session_id: str):
        body_args = deserialize_serializable(self.request.body)
        storage_api_to_gets = defaultdict(list)
        storage_api_to_idx = defaultdict(list)
        results = [None] * len(body_args)
        for i, (data_key, conditions, error) in enumerate(body_args):
            oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
            storage_api_to_idx[oscar_api].append(i)
            storage_api_to_gets[oscar_api].append(
                oscar_api.get.delay(data_key, conditions=conditions, error=error)
            )
        for api, fetches in storage_api_to_gets.items():
            data_list = await api.get.batch(*fetches)
            for idx, data in zip(storage_api_to_idx[api], data_list):
                results[idx] = data
        res_data = serialize_serializable(results)
        self.write(res_data)

    @web_api("(?P<data_key>[^/]+)", method="post")
    async def get_data_by_post(self, session_id: str, data_key: str):
        body_args = (
            deserialize_serializable(self.request.body) if self.request.body else None
        )
        conditions = body_args.get("conditions")

        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        result = await oscar_api.get(data_key, conditions)
        self.write(serialize_serializable(result))

    @web_api("(?P<data_key>[^/]+)", method="put")
    async def put_data(self, session_id: str, data_key: str):
        level = self.get_argument("level", None) or "MEMORY"
        level = getattr(StorageLevel, level.upper())

        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        res = await oscar_api.put(
            data_key, deserialize_serializable(self.request.body), level
        )
        self.write(serialize_serializable(res))

    @web_api("(?P<data_key>[^/]+)", method="get", arg_filter={"action": "get_infos"})
    async def get_infos(self, session_id: str, data_key: str):
        oscar_api = await self._get_storage_api_by_object_id(session_id, data_key)
        res = await oscar_api.get_infos(data_key)
        self.write(serialize_serializable(res))


web_handlers = {StorageWebAPIHandler.get_root_pattern(): StorageWebAPIHandler}


class WebStorageAPI(AbstractStorageAPI, MarsWebAPIClientMixin):
    def __init__(
        self,
        session_id: str,
        address: str,
        band_name: str,
        request_rewriter: Callable = None,
    ):
        self._session_id = session_id
        self._address = address.rstrip("/")
        self._band_name = band_name
        self.request_rewriter = request_rewriter

    @mo.extensible
    async def get(
        self, data_key: str, conditions: List = None, error: str = "raise"
    ) -> Any:
        path = f"{self._address}/api/session/{self._session_id}/storage/{data_key}"
        params = dict(error=error)
        if conditions is not None:
            params["conditions"] = conditions
        body = serialize_serializable(params)
        res = await self._request_url(
            path=path,
            method="POST",
            headers={"Content-Type": "application/octet-stream"},
            data=body,
        )
        return deserialize_serializable(res.body)

    @get.batch
    async def get_batch(self, args_list, kwargs_list):
        get_chunks = []
        for args, kwargs in zip(args_list, kwargs_list):
            data_key, conditions, error = self.get.bind(*args, **kwargs)
            get_chunks.append([data_key, conditions, error])

        path = f"{self._address}/api/session/{self._session_id}/storage/batch/get"
        res = await self._request_url(
            path=path,
            method="POST",
            data=serialize_serializable(get_chunks),
        )
        return deserialize_serializable(res.body)

    @mo.extensible
    async def put(
        self, data_key: str, obj: object, level: StorageLevel = StorageLevel.MEMORY
    ) -> DataInfo:
        params = dict(level=level.name.lower())
        path = f"{self._address}/api/session/{self._session_id}/storage/{data_key}"
        res = await self._request_url(
            path=path,
            method="PUT",
            params=params,
            headers={"Content-Type": "application/octet-stream"},
            data=serialize_serializable(obj),
        )
        return deserialize_serializable(res.body)

    @mo.extensible
    async def get_infos(self, data_key: str) -> List[DataInfo]:
        path = f"{self._address}/api/session/{self._session_id}/storage/{data_key}"
        params = dict(action="get_infos")
        res = await self._request_url(
            path=path,
            method="GET",
            headers={"Content-Type": "application/octet-stream"},
            params=params,
        )
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils import check_random_state
from ..datasource import tensor as astensor
from .lu import lu
from .qr import qr
from .svd import svd
from .utils import svd_flip


# ---------------------------------------------------------------------
# Original implementation is in `sklearn.utils.extmath.randomized_svd`.
# ---------------------------------------------------------------------


def randomized_range_finder(
    A, size, n_iter, power_iteration_normalizer="auto", random_state=None
):
    r"""Computes an orthonormal matrix whose range approximates the range of A.

    .. versionadded:: 0.1.3

    Parameters
    ----------
    A : 2D tensor
        The input data tensor

    size : integer
        Size of the return tensor

    n_iter : integer
        Number of power iterations used to stabilize the result

    power_iteration_normalizer : 'auto' (default), 'QR', 'LU', 'none'
        Whether the power iterations are normalized with step-by-step
        QR factorization (the slowest but most accurate), 'none'
        (the fastest but numerically unstable when `n_iter` is large, e.g.
        typically 5 or larger), or 'LU' factorization (numerically stable
        but can lose slightly in accuracy). The 'auto' mode applies no
        normalization if `n_iter` <= 2 and switches to LU otherwise.

    random_state : int, RandomState instance or None, optional (default=None)
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`.

    Returns
    -------
    Q : 2D array
        A (size x size) projection matrix, the range of which
        approximates well the range of the input matrix A.

    Notes
    -----

    Follows Algorithm 4.3 of
    Finding structure with randomness: Stochastic algorithms for constructing
    approximate matrix decompositions
    Halko, et al., 2009 (arXiv:909) https://arxiv.org/pdf/0909.4061.pdf

    An implementation of a randomized algorithm for principal component
    analysis
    A. Szlam et al. 2014
    """
    random_state = check_random_state(random_state)

    # Generating normal random vectors with shape: (A.shape[1], size)
    Q = random_state.normal(size=(A.shape[1], size))
    if A.dtype.kind == "f":
        # Ensure f32 is preserved as f32
        Q = Q.astype(A.dtype, copy=False)

    # Deal with "auto" mode
    if power_iteration_normalizer == "auto":
        if n_iter <= 2:
            power_iteration_normalizer = "none"
        else:
            power_iteration_normalizer = "LU"

    # Perform power iterations with Q to further 'imprint' the top
    # singular vectors of A in Q
    for _ in range(n_iter):
        if power_iteration_normalizer == "none":
            Q = A.dot(Q)
            Q = A.T.dot(Q)
        elif power_iteration_normalizer == "LU":
            # TODO: directly get Q when lu supports `permute_l`
            p, l, _ = lu(A.dot(Q))
            Q = p.dot(l)
            p, l, _ = lu(A.T.dot(Q))
            Q = p.dot(l)
        elif power_iteration_normalizer == "QR":
            Q, _ = qr(A.dot(Q))
            Q, _ = qr(A.T.dot(Q))

    # Sample the range of A using by linear projection of Q
    # Extract an orthonormal basis
    Q, _ = qr(A.dot(Q))
    return Q


def randomized_svd(
    M,
    n_components,
    n_oversamples=10,
    n_iter="auto",
    power_iteration_normalizer="auto",
    transpose="auto",
    flip_sign=True,
    random_state=0,
):
    r"""
    Computes a truncated randomized SVD

    .. versionadded:: 0.1.4

    Parameters
    ----------
    M : Tensor
        tensor to decompose
    n_components : int
        Number of singular values and vectors to extract.
    n_oversamples : int (default is 10)
        Additional number of random vectors to sample the range of M so as
        to ensure proper conditioning. The total number of random vectors
        used to find the range of M is n_components + n_oversamples. Smaller
        number can improve speed but can negatively impact the quality of
        approximation of singular vectors and singular values.
    n_iter : int or 'auto' (default is 'auto')
        Number of power iterations. It can be used to deal with very noisy
        problems. When 'auto', it is set to 4, unless `n_components` is small
        (< .1 * min(X.shape)) `n_iter` in which case is set to 7.
        This improves precision with few components.
    power_iteration_normalizer : 'auto' (default), 'QR', 'LU', 'none'
        Whether the power iterations are normalized with step-by-step
        QR factorization (the slowest but most accurate), 'none'
        (the fastest but numerically unstable when `n_iter` is large, e.g.
        typically 5 or larger), or 'LU' factorization (numerically stable
        but can lose slightly in accuracy). The 'auto' mode applies no
        normalization if `n_iter` <= 2 and switches to LU otherwise.
    transpose : True, False or 'auto' (default)
        Whether the algorithm should be applied to M.T instead of M. The
        result should approximately be the same. The 'auto' mode will
        trigger the transposition if M.shape[1] > M.shape[0] since this
        implementation of randomized SVD tend to be a little faster in that
        case.
    flip_sign : boolean, (True by default)
        The output of a singular value decomposition is only unique up to a
        permutation of the signs of the singular vectors. If `flip_sign` is
        set to `True`, the sign ambiguity is resolved by making the largest
        loadings for each component in the left singular vectors positive.
    random_state : int, RandomState instance or None, optional (default=None)
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`.
    Notes
    -----
    This algorithm finds a (usually very good) approximate truncated
    singular value decomposition using randomization to speed up the
    computations. It is particularly fast on large matrices on which
    you wish to extract only a small number of components. In order to
    obtain further speed up, `n_iter` can be set <=2 (at the cost of
    loss of precision).
    References
    ----------
    * Finding structure with randomness: Stochastic algorithms for constructing
      approximate matrix decompositions
      Halko, et al., 2009 https://arxiv.org/abs/0909.4061
    * A randomized algorithm for the decomposition of matrices
      Per-Gunnar Martinsson, Vladimir Rokhlin and Mark Tygert
    * An implementation of a randomized algorithm for principal component
      analysis
      A. Szlam et al. 2014
    """
    M = astensor(M)
    random_state = check_random_state(random_state)
    n_random = n_components + n_oversamples
    n_samples, n_features = M.shape

    if n_iter == "auto":
        # Check if the number of iterations is explicitly specified
        # Adjust n_iter. 7 was found a good compromise for PCA.
        # https://github.com/scikit-learn/scikit-learn/pull/5299
        n_iter = 7 if n_components < 0.1 * min(M.shape) else 4

    if transpose == "auto":
        transpose = n_samples < n_features
    if transpose:
        # this implementation is a bit faster with smaller shape[1]
        M = M.T

    Q = randomized_range_finder(
        M, n_random, n_iter, power_iteration_normalizer, random_state
    )
    # project M to the (k + p) dimensional space using the basis vectors
    B = Q.T.dot(M)

    # compute the SVD on the thin matrix: (k + p) wide
    Uhat, s, V = svd(B)

    U = Q.dot(Uhat)

    if flip_sign:
        if not transpose:
            U, V = svd_flip(U, V)
        else:
            # In case of transpose u_based_decision=false
            # to actually flip based on u and not v.
            U, V = svd_flip(U, V, u_based_decision=False)

    if transpose:
        # transpose back the results according to the input convention
        return V[:n_components, :].T, s[:n_components], U[:, :n_components].T
    else:
        return U[:, :n_components], s[:n_components], V[:n_components, :]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import Float64Field, BoolField
from ..array_utils import device, as_same_device
from .core import TensorBinOp


class TensorIsclose(TensorBinOp):
    _op_type_ = OperandDef.ISCLOSE

    _rtol = Float64Field("rtol")
    _atol = Float64Field("atol")
    _equal_nan = BoolField("equal_nan")

    def __init__(
        self,
        rtol=None,
        atol=None,
        equal_nan=None,
        casting="same_kind",
        err=None,
        sparse=False,
        **kw
    ):
        err = err if err is not None else np.geterr()
        super().__init__(
            _rtol=rtol,
            _atol=atol,
            _equal_nan=equal_nan,
            _casting=casting,
            _err=err,
            sparse=sparse,
            **kw
        )

    @property
    def rtol(self):
        return self._rtol

    @property
    def atol(self):
        return self._atol

    @property
    def equal_nan(self):
        return self._equal_nan

    @classmethod
    def _is_sparse(cls, x1, x2):
        if (
            hasattr(x1, "issparse")
            and x1.issparse()
            and np.isscalar(x2)
            and not np.isclose(x2, 0)
        ):
            return True
        if (
            hasattr(x2, "issparse")
            and x2.issparse()
            and np.isscalar(x1)
            and not np.isclose(x1, 0)
        ):
            return True
        return False

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            a = op.lhs if np.isscalar(op.lhs) else inputs[0]
            b = op.rhs if np.isscalar(op.rhs) else inputs[-1]

            ctx[op.outputs[0].key] = xp.isclose(
                a, b, atol=op.atol, rtol=op.rtol, equal_nan=op.equal_nan
            )


def isclose(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
    """
    Returns a boolean tensor where two tensors are element-wise equal within a
    tolerance.

    The tolerance values are positive, typically very small numbers.  The
    relative difference (`rtol` * abs(`b`)) and the absolute difference
    `atol` are added together to compare against the absolute difference
    between `a` and `b`.

    Parameters
    ----------
    a, b : array_like
        Input tensors to compare.
    rtol : float
        The relative tolerance parameter (see Notes).
    atol : float
        The absolute tolerance parameter (see Notes).
    equal_nan : bool
        Whether to compare NaN's as equal.  If True, NaN's in `a` will be
        considered equal to NaN's in `b` in the output tensor.

    Returns
    -------
    y : array_like
        Returns a boolean tensor of where `a` and `b` are equal within the
        given tolerance. If both `a` and `b` are scalars, returns a single
        boolean value.

    See Also
    --------
    allclose

    Notes
    -----

    For finite values, isclose uses the following equation to test whether
    two floating point values are equivalent.

     absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))

    The above equation is not symmetric in `a` and `b`, so that
    `isclose(a, b)` might be different from `isclose(b, a)` in
    some rare cases.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.isclose([1e10,1e-7], [1.00001e10,1e-8]).execute()
    array([True, False])
    >>> mt.isclose([1e10,1e-8], [1.00001e10,1e-9]).execute()
    array([True, True])
    >>> mt.isclose([1e10,1e-8], [1.0001e10,1e-9]).execute()
    array([False, True])
    >>> mt.isclose([1.0, mt.nan], [1.0, mt.nan]).execute()
    array([True, False])
    >>> mt.isclose([1.0, mt.nan], [1.0, mt.nan], equal_nan=True).execute()
    array([True, True])
    """
    op = TensorIsclose(rtol=rtol, atol=atol, equal_nan=equal_nan, dtype=np.dtype(bool))
    return op(a, b)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...lib.filesystem import open_file
from .core import TensorFromHDF5Like, TensorOrder


class TensorHDF5DataSource(TensorFromHDF5Like):
    _op_type_ = OperandDef.TENSOR_FROM_HDF5

    @classmethod
    def execute(cls, ctx, op):
        import h5py

        axis_offsets = op.axis_offsets
        shape = op.outputs[0].shape

        with h5py.File(open_file(op.filename), mode="r") as f:
            ds = f[op.path]
            data = ds[
                tuple(
                    slice(offset, offset + size)
                    for offset, size in zip(axis_offsets, shape)
                )
            ]
            ctx[op.outputs[0].key] = data


def fromhdf5(hdf5_file, group=None, dataset=None, chunk_size=None):
    import h5py

    if isinstance(hdf5_file, h5py.Dataset):
        filename = hdf5_file.file.filename
        group = hdf5_file.parent.name
        dataset = hdf5_file.name.rsplit("/", 1)[1]
        chunk_size = chunk_size if chunk_size is not None else hdf5_file.chunks
        shape = hdf5_file.shape
        dtype = hdf5_file.dtype
    elif isinstance(hdf5_file, h5py.File):
        filename = hdf5_file.filename
        if dataset is None:
            raise ValueError("`dataset` should be provided")
        try:
            h5_dataset = hdf5_file[TensorHDF5DataSource.get_path(group, dataset)]
        except KeyError:
            raise ValueError(f"dataset({dataset}) does not exist")
        chunk_size = chunk_size if chunk_size is not None else h5_dataset.chunks
        shape = h5_dataset.shape
        dtype = h5_dataset.dtype
    elif isinstance(hdf5_file, str):
        filename = hdf5_file
        try:
            with h5py.File(open_file(filename), mode="r") as f:
                if dataset is None:
                    raise ValueError("`dataset` should be provided")
                h5_dataset = f[TensorHDF5DataSource.get_path(group, dataset)]

                chunk_size = chunk_size if chunk_size is not None else h5_dataset.chunks
                shape = h5_dataset.shape
                dtype = h5_dataset.dtype
        except KeyError:
            raise ValueError(f"dataset({dataset}) does not exist")
    else:
        raise TypeError(
            "`hdf5_file` passed has wrong type, "
            "expect str, h5py.File or h5py.Dataset, "
            f"got {type(hdf5_file)}"
        )

    op = TensorHDF5DataSource(
        filename=filename, group=group, dataset=dataset, dtype=dtype
    )
    return op(shape, chunk_size=chunk_size, order=TensorOrder.C_ORDER)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools
import random
from collections.abc import Iterable

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import ENTITY_TYPE, OutputType, get_output_types, recursive_tile
from ...core.operand import OperandStage, MapReduceOperand
from ...serialization.serializables import (
    BoolField,
    DictField,
    Float32Field,
    KeyField,
    Int32Field,
    Int64Field,
    NDArrayField,
    StringField,
)
from ...tensor.operands import TensorShuffleProxy
from ...tensor.random import RandomStateField
from ...tensor.utils import gen_random_seeds
from ...utils import has_unknown_shape
from ..initializer import Series as asseries
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import parse_index

_ILOC_COL_HEADER = "_gsamp_iloc_col_"
_WEIGHT_COL_HEADER = "_gsamp_weight_col_"


# code adapted from pandas.core.groupby.groupby.DataFrameGroupBy.sample
def _sample_groupby_iter(
    groupby, obj_index, n, frac, replace, weights, random_state=None, errors="ignore"
):
    if weights is None:
        ws = [None] * groupby.ngroups
    elif not isinstance(weights, Iterable) or isinstance(weights, str):
        ws = [weights] * groupby.ngroups
    else:
        weights = pd.Series(weights, index=obj_index)
        ws = [weights.iloc[idx] for idx in groupby.indices.values()]

    group_iterator = groupby.grouper.get_iterator(groupby._selected_obj)
    if not replace and errors == "ignore":
        for (_, obj), w in zip(group_iterator, ws):
            yield obj.sample(
                n=n, frac=frac, replace=replace, weights=w, random_state=random_state
            ) if len(obj) > n else obj
    else:
        for (_, obj), w in zip(group_iterator, ws):
            yield obj.sample(
                n=n, frac=frac, replace=replace, weights=w, random_state=random_state
            )


class GroupBySampleILoc(DataFrameOperand, DataFrameOperandMixin):
    _op_code_ = opcodes.GROUPBY_SAMPLE_ILOC
    _op_module_ = "dataframe.groupby"

    _groupby_params = DictField("groupby_params")
    _size = Int64Field("size")
    _frac = Float32Field("frac")
    _replace = BoolField("replace")
    _weights = KeyField("weights")
    _seed = Int32Field("seed")
    _random_state = RandomStateField("random_state")
    _errors = StringField("errors")

    _random_col_id = Int32Field("random_col_id")

    # for chunks
    # num of instances for chunks
    _left_iloc_bound = Int64Field("left_iloc_bound")

    def __init__(
        self,
        groupby_params=None,
        size=None,
        frac=None,
        replace=None,
        weights=None,
        random_state=None,
        seed=None,
        errors=None,
        left_iloc_bound=None,
        random_col_id=None,
        **kw
    ):
        super().__init__(
            _groupby_params=groupby_params,
            _size=size,
            _frac=frac,
            _seed=seed,
            _replace=replace,
            _weights=weights,
            _random_state=random_state,
            _errors=errors,
            _left_iloc_bound=left_iloc_bound,
            _random_col_id=random_col_id,
            **kw
        )
        if self._random_col_id is None:
            self._random_col_id = random.randint(10000, 99999)

    @property
    def groupby_params(self):
        return self._groupby_params

    @property
    def size(self):
        return self._size

    @property
    def frac(self):
        return self._frac

    @property
    def replace(self):
        return self._replace

    @property
    def weights(self):
        return self._weights

    @property
    def seed(self):
        return self._seed

    @property
    def random_state(self):
        if self._random_state is None:
            self._random_state = np.random.RandomState(self.seed)
        return self._random_state

    @property
    def errors(self):
        return self._errors

    @property
    def left_iloc_bound(self):
        return self._left_iloc_bound

    @property
    def random_col_id(self):
        return self._random_col_id

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        input_iter = iter(inputs)
        next(input_iter)
        if isinstance(self.weights, ENTITY_TYPE):
            self._weights = next(input_iter)

    def __call__(self, df):
        self._output_types = [OutputType.tensor]
        inp_tileables = [df]
        if self.weights is not None:
            inp_tileables.append(self.weights)
        return self.new_tileable(
            inp_tileables, dtype=np.dtype(np.int_), shape=(np.nan,)
        )

    @classmethod
    def tile(cls, op: "GroupBySampleILoc"):
        in_df = op.inputs[0]
        out_tensor = op.outputs[0]
        iloc_col_header = _ILOC_COL_HEADER + str(op.random_col_id)
        weight_col_header = _WEIGHT_COL_HEADER + str(op.random_col_id)

        if has_unknown_shape(in_df):
            yield

        if op.weights is None:
            weights_iter = itertools.repeat(None)
        else:
            weights_iter = iter(op.weights.chunks)

        if isinstance(op.groupby_params["by"], list):
            map_cols = list(op.groupby_params["by"])
        else:  # pragma: no cover
            map_cols = []

        dtypes = in_df.dtypes.copy()
        dtypes.at[iloc_col_header] = np.dtype(np.int_)
        map_cols.append(iloc_col_header)
        if op.weights is not None:
            dtypes.at[weight_col_header] = op.weights.dtype
            map_cols.append(weight_col_header)

        new_dtypes = dtypes[map_cols]
        new_columns_value = parse_index(new_dtypes.index, store_data=True)

        map_chunks = []
        left_ilocs = np.array((0,) + in_df.nsplits[0]).cumsum()
        for inp_chunk, weight_chunk in zip(in_df.chunks, weights_iter):
            new_op = op.copy().reset_key()
            new_op._left_iloc_bound = int(left_ilocs[inp_chunk.index[0]])
            new_op.stage = OperandStage.map
            new_op._output_types = [OutputType.dataframe]

            inp_chunks = [inp_chunk]
            if weight_chunk is not None:
                inp_chunks.append(weight_chunk)
            params = inp_chunk.params
            params.update(
                dict(
                    dtypes=new_dtypes,
                    columns_value=new_columns_value,
                    shape=(inp_chunk.shape[0], len(new_dtypes)),
                    index=inp_chunk.index,
                )
            )
            map_chunks.append(new_op.new_chunk(inp_chunks, **params))

        new_op = op.copy().reset_key()
        new_op._output_types = [OutputType.dataframe]
        params = in_df.params
        params.update(
            dict(
                chunks=map_chunks,
                nsplits=(in_df.nsplits[0], (len(new_dtypes),)),
                dtypes=new_dtypes,
                columns_value=new_columns_value,
                shape=(in_df.shape[0], len(new_dtypes)),
            )
        )
        map_df = new_op.new_tileable(op.inputs, **params)

        groupby_params = op.groupby_params.copy()
        groupby_params.pop("selection", None)
        grouped = yield from recursive_tile(map_df.groupby(**groupby_params))

        result_chunks = []
        seeds = gen_random_seeds(len(grouped.chunks), op.random_state)
        for group_chunk, seed in zip(grouped.chunks, seeds):
            new_op = op.copy().reset_key()
            new_op.stage = OperandStage.reduce
            new_op._weights = None
            new_op._random_state = None
            new_op._seed = seed

            result_chunks.append(
                new_op.new_chunk(
                    [group_chunk],
                    shape=(np.nan,),
                    index=(group_chunk.index[0],),
                    dtype=out_tensor.dtype,
                )
            )

        new_op = op.copy().reset_key()
        params = out_tensor.params
        params.update(
            dict(chunks=result_chunks, nsplits=((np.nan,) * len(result_chunks),))
        )
        return new_op.new_tileables(op.inputs, **params)

    @classmethod
    def execute(cls, ctx, op: "GroupBySampleILoc"):
        in_data = ctx[op.inputs[0].key]
        iloc_col = _ILOC_COL_HEADER + str(op.random_col_id)
        weight_col = _WEIGHT_COL_HEADER + str(op.random_col_id)
        if op.stage == OperandStage.map:
            if op.weights is not None:
                ret = pd.DataFrame(
                    {
                        iloc_col: np.arange(
                            op.left_iloc_bound, op.left_iloc_bound + len(in_data)
                        ),
                        weight_col: ctx[op.weights.key],
                    },
                    index=in_data.index,
                )
            else:
                ret = pd.DataFrame(
                    {
                        iloc_col: np.arange(
                            op.left_iloc_bound, op.left_iloc_bound + len(in_data)
                        ),
                    },
                    index=in_data.index,
                )

            if isinstance(op.groupby_params["by"], list):
                ret = pd.concat([in_data[op.groupby_params["by"]], ret], axis=1)

            ctx[op.outputs[0].key] = ret
        else:
            if weight_col not in in_data.obj.columns:
                weight_col = None

            if len(in_data.obj) == 0 or in_data.ngroups == 0:
                ctx[op.outputs[0].key] = np.array([], dtype=np.int_)
            else:
                ctx[op.outputs[0].key] = np.concatenate(
                    [
                        sample_pd[iloc_col].to_numpy()
                        for sample_pd in _sample_groupby_iter(
                            in_data,
                            in_data.obj.index,
                            n=op.size,
                            frac=op.frac,
                            replace=op.replace,
                            weights=weight_col,
                            random_state=op.random_state,
                            errors=op.errors,
                        )
                    ]
                )


class GroupBySample(MapReduceOperand, DataFrameOperandMixin):
    _op_code_ = opcodes.RAND_SAMPLE
    _op_module_ = "dataframe.groupby"

    _groupby_params = DictField("groupby_params")
    _size = Int64Field("size")
    _frac = Float32Field("frac")
    _replace = BoolField("replace")
    _weights = KeyField("weights")
    _seed = Int32Field("seed")
    _random_state = RandomStateField("random_state")
    _errors = StringField("errors")

    # for chunks
    # num of instances for chunks
    _input_nsplits = NDArrayField("input_nsplits")

    def __init__(
        self,
        groupby_params=None,
        size=None,
        frac=None,
        replace=None,
        weights=None,
        random_state=None,
        seed=None,
        errors=None,
        input_nsplits=None,
        **kw
    ):
        super().__init__(
            _groupby_params=groupby_params,
            _size=size,
            _frac=frac,
            _seed=seed,
            _replace=replace,
            _weights=weights,
            _random_state=random_state,
            _errors=errors,
            _input_nsplits=input_nsplits,
            **kw
        )

    @property
    def groupby_params(self):
        return self._groupby_params

    @property
    def size(self):
        return self._size

    @property
    def frac(self):
        return self._frac

    @property
    def replace(self):
        return self._replace

    @property
    def weights(self):
        return self._weights

    @property
    def seed(self):
        return self._seed

    @property
    def random_state(self):
        return self._random_state

    @property
    def errors(self):
        return self._errors

    @property
    def input_nsplits(self):
        return self._input_nsplits

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        input_iter = iter(inputs)
        next(input_iter)
        if isinstance(self.weights, ENTITY_TYPE):
            self._weights = next(input_iter)

    def __call__(self, groupby):
        df = groupby
        while df.op.output_types[0] not in (OutputType.dataframe, OutputType.series):
            df = df.inputs[0]

        selection = groupby.op.groupby_params.pop("selection", None)
        if df.ndim > 1 and selection:
            if isinstance(selection, tuple) and selection not in df.dtypes:
                selection = list(selection)
            result_df = df[selection]
        else:
            result_df = df

        params = result_df.params
        params["shape"] = (
            (np.nan,) if result_df.ndim == 1 else (np.nan, result_df.shape[-1])
        )
        params["index_value"] = parse_index(result_df.index_value.to_pandas()[:0])

        input_dfs = [df]
        if isinstance(self.weights, ENTITY_TYPE):
            input_dfs.append(self.weights)

        self._output_types = get_output_types(result_df)
        return self.new_tileable(input_dfs, **params)

    @classmethod
    def _tile_one_chunk(cls, op: "GroupBySample", in_df, weights):
        out = op.outputs[0]

        input_dfs = [in_df]
        if isinstance(weights, ENTITY_TYPE):
            input_dfs.append(weights)

        params = out.params
        chunk_op = op.copy().reset_key()
        if isinstance(weights, ENTITY_TYPE):
            chunk_op._weights = weights
        params["index"] = (0,) * out.ndim
        chunk = chunk_op.new_chunk([c.chunks[0] for c in input_dfs], **params)

        df_op = op.copy().reset_key()
        return df_op.new_tileables(
            input_dfs, chunks=[chunk], nsplits=((s,) for s in out.shape), **params
        )

    @classmethod
    def _tile_distributed(cls, op: "GroupBySample", in_df, weights):
        out_df = op.outputs[0]
        if has_unknown_shape(in_df):
            yield

        sample_iloc_op = GroupBySampleILoc(
            groupby_params=op.groupby_params,
            size=op.size,
            frac=op.frac,
            replace=op.replace,
            weights=weights,
            random_state=op.random_state,
            errors=op.errors,
            seed=None,
            left_iloc_bound=None,
        )
        sampled_iloc = yield from recursive_tile(sample_iloc_op(in_df))

        map_chunks = []
        for c in sampled_iloc.chunks:
            new_op = op.copy().reset_key()
            new_op.stage = OperandStage.map
            new_op._weights = None
            new_op._output_types = [OutputType.tensor]
            new_op._input_nsplits = np.array(in_df.nsplits[0])

            map_chunks.append(
                new_op.new_chunk(
                    [c], dtype=sampled_iloc.dtype, shape=(np.nan,), index=c.index
                )
            )

        proxy_chunk = TensorShuffleProxy(dtype=sampled_iloc.dtype).new_chunk(
            map_chunks, shape=()
        )

        reduce_chunks = []
        for ordinal, src_chunk in enumerate(in_df.chunks):
            new_op = op.copy().reset_key()
            new_op._weights = None
            new_op._output_types = [OutputType.tensor]
            new_op.stage = OperandStage.reduce
            new_op.reducer_index = (src_chunk.index[0],)
            new_op.reducer_ordinal = ordinal
            new_op.n_reducers = len(in_df.chunks)
            new_op._input_nsplits = np.array(in_df.nsplits[0])

            reduce_chunks.append(
                new_op.new_chunk(
                    [proxy_chunk],
                    index=src_chunk.index,
                    dtype=sampled_iloc.dtype,
                    shape=(np.nan,),
                )
            )

        combine_chunks = []
        for src_chunk, reduce_chunk in zip(in_df.chunks, reduce_chunks):
            new_op = op.copy().reset_key()
            new_op.stage = OperandStage.combine
            new_op._weights = None

            params = out_df.params
            if out_df.ndim == 2:
                params.update(
                    dict(
                        index=src_chunk.index,
                        dtypes=out_df.dtypes,
                        shape=(np.nan, out_df.shape[1]),
                        columns_value=out_df.columns_value,
                    )
                )
            else:
                params.update(
                    dict(
                        index=(src_chunk.index[0],),
                        dtype=out_df.dtype,
                        shape=(np.nan,),
                        name=out_df.name,
                    )
                )
            combine_chunks.append(new_op.new_chunk([src_chunk, reduce_chunk], **params))

        new_op = op.copy().reset_key()
        if out_df.ndim == 2:
            new_nsplits = ((np.nan,) * in_df.chunk_shape[0], (out_df.shape[1],))
        else:
            new_nsplits = ((np.nan,) * in_df.chunk_shape[0],)
        return new_op.new_tileables(
            out_df.inputs, chunks=combine_chunks, nsplits=new_nsplits, **out_df.params
        )

    @classmethod
    def tile(cls, op: "GroupBySample"):
        in_df = op.inputs[0]
        if in_df.ndim == 2:
            in_df = yield from recursive_tile(in_df.rechunk({1: (in_df.shape[1],)}))

        weights = op.weights
        if isinstance(weights, ENTITY_TYPE):
            weights = yield from recursive_tile(weights.rechunk({0: in_df.nsplits[0]}))

        if len(in_df.chunks) == 1:
            return cls._tile_one_chunk(op, in_df, weights)
        return (yield from cls._tile_distributed(op, in_df, weights))

    @classmethod
    def execute(cls, ctx, op: "GroupBySample"):
        out_df = op.outputs[0]

        if op.stage == OperandStage.map:
            in_data = ctx[op.inputs[0].key]
            in_data = np.sort(in_data)
            input_nsplits = np.copy(op.input_nsplits).tolist()
            pos_array = np.cumsum([0] + input_nsplits)
            poses = np.searchsorted(in_data, pos_array).tolist()
            for idx, (left, right) in enumerate(zip(poses, poses[1:])):
                ctx[op.outputs[0].key, (idx,)] = in_data[left:right]
        elif op.stage == OperandStage.reduce:
            in_indexes = list(op.iter_mapper_data(ctx))
            idx = np.sort(np.concatenate(in_indexes))
            if op.outputs[0].index[0] > 0:
                acc_nsplits = np.cumsum(op.input_nsplits)
                idx -= acc_nsplits[op.outputs[0].index[0] - 1]
            ctx[op.outputs[0].key] = idx
        elif op.stage == OperandStage.combine:
            in_data = ctx[op.inputs[0].key]
            idx = ctx[op.inputs[1].key]
            selection = op.groupby_params.get("selection")
            if selection:
                in_data = in_data[selection]
            ctx[op.outputs[0].key] = in_data.iloc[idx]
        else:
            in_data = ctx[op.inputs[0].key]
            weights = op.weights
            if isinstance(weights, ENTITY_TYPE):
                weights = ctx[weights.key]
            params = op.groupby_params.copy()
            selection = params.pop("selection", None)

            grouped = in_data.groupby(**params)
            if selection is not None:
                grouped = grouped[selection]

            result = pd.concat(
                [
                    sample_df
                    for sample_df in _sample_groupby_iter(
                        grouped,
                        in_data.index,
                        n=op.size,
                        frac=op.frac,
                        replace=op.replace,
                        weights=weights,
                        random_state=op.random_state,
                        errors=op.errors,
                    )
                ]
            )
            ctx[out_df.key] = result


def groupby_sample(
    groupby,
    n=None,
    frac=None,
    replace=False,
    weights=None,
    random_state=None,
    errors="ignore",
):
    """
    Return a random sample of items from each group.

    You can use `random_state` for reproducibility.

    Parameters
    ----------
    n : int, optional
        Number of items to return for each group. Cannot be used with
        `frac` and must be no larger than the smallest group unless
        `replace` is True. Default is one if `frac` is None.
    frac : float, optional
        Fraction of items to return. Cannot be used with `n`.
    replace : bool, default False
        Allow or disallow sampling of the same row more than once.
    weights : list-like, optional
        Default None results in equal probability weighting.
        If passed a list-like then values must have the same length as
        the underlying DataFrame or Series object and will be used as
        sampling probabilities after normalization within each group.
        Values must be non-negative with at least one positive element
        within each group.
    random_state : int, array-like, BitGenerator, np.random.RandomState, optional
        If int, array-like, or BitGenerator (NumPy>=1.17), seed for
        random number generator
        If np.random.RandomState, use as numpy RandomState object.
    errors : {'ignore', 'raise'}, default 'ignore'
        If ignore, errors will not be raised when `replace` is False
        and size of some group is less than `n`.

    Returns
    -------
    Series or DataFrame
        A new object of same type as caller containing items randomly
        sampled within each group from the caller object.

    See Also
    --------
    DataFrame.sample: Generate random samples from a DataFrame object.
    numpy.random.choice: Generate a random sample from a given 1-D numpy
        array.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame(
    ...     {"a": ["red"] * 2 + ["blue"] * 2 + ["black"] * 2, "b": range(6)}
    ... )
    >>> df.execute()
           a  b
    0    red  0
    1    red  1
    2   blue  2
    3   blue  3
    4  black  4
    5  black  5

    Select one row at random for each distinct value in column a. The
    `random_state` argument can be used to guarantee reproducibility:

    >>> df.groupby("a").sample(n=1, random_state=1).execute()
           a  b
    4  black  4
    2   blue  2
    1    red  1

    Set `frac` to sample fixed proportions rather than counts:

    >>> df.groupby("a")["b"].sample(frac=0.5, random_state=2).execute()
    5    5
    2    2
    0    0
    Name: b, dtype: int64

    Control sample probabilities within groups by setting weights:

    >>> df.groupby("a").sample(
    ...     n=1,
    ...     weights=[1, 1, 1, 0, 0, 1],
    ...     random_state=1,
    ... ).execute()
           a  b
    5  black  5
    2   blue  2
    0    red  0
    """
    groupby_params = groupby.op.groupby_params.copy()
    groupby_params.pop("as_index", None)

    if weights is not None and not isinstance(weights, ENTITY_TYPE):
        weights = asseries(weights)

    n = 1 if n is None and frac is None else n
    rs = copy.deepcopy(
        random_state.to_numpy() if hasattr(random_state, "to_numpy") else random_state
    )
    op = GroupBySample(
        size=n,
        frac=frac,
        replace=replace,
        weights=weights,
        random_state=rs,
        groupby_params=groupby_params,
        errors=errors,
    )
    return op(groupby)
